__all__ = ("cvs_syncer",)

import os

from . import base


class cvs_syncer(base.VcsSyncer):

    binary = "cvs"

    supported_uris = (
        ('cvs+', 5),
        ('cvs://', 5),
    )

    @classmethod
    def is_usable_on_filepath(cls, path):
        cvs_path = os.path.join(path, "CVS")
        if cls.disabled or not os.path.isdir(cvs_path):
            return None
        return (cls._rewrite_uri_from_stat(cvs_path, 'cvs://'),)

    @classmethod
    def _parse_uri(cls, raw_uri):
        if not raw_uri.startswith("cvs") and \
                not raw_uri.startswith("cvs+"):
            raise base.UriError(raw_uri, "must be cvs:// or cvs+${RSH}")
        if raw_uri.startswith("cvs://"):
            return None, raw_uri[len("cvs://"):]
        proto = raw_uri[len("cvs+"):].split(":", 1)
        if not proto[0]:
            raise base.UriError(
                raw_uri, "cvs+ requires the rsh alternative to be specified")
        if proto[0] == "anon":
            proto[0] = None
        elif proto[0] != "pserver":
            try:
                proto[0] = cls.require_binary(proto[0])
            except base.MissingBinary:
                raise base.UriError(
                    raw_uri, f"missing rsh binary: {proto[0]!r}")
        return proto[0], proto[1].lstrip("/")

    def __init__(self, basedir, raw_uri, **kwargs):
        proto, uri = self._parse_uri(raw_uri)
        self.rsh = proto
        if self.rsh is None:
            uri = f":anoncvs:{uri}"
        elif self.rsh == "pserver":
            uri = f":pserver:{uri}"
            self.rsh = None
        else:
            uri = f":ext:{uri}"
        host, self.module = uri.rsplit(":", 1)
        super().__init__(basedir, host, **kwargs)

        self.env['CVSROOT'] = self.uri
        if self.rsh is not None:
            self.env['CVS_RSH'] = self.rsh

    def _update_existing(self):
        return [self.binary_path, "up"]

    def _initial_pull(self):
        return [self.binary_path, "co", "-d", self.basedir]



__all__ = ("PigeonHoledSlots",)

from ..restrictions import restriction

# lil too getter/setter like for my tastes...


class PigeonHoledSlots:
    """class for tracking slotting to a specific atom/obj key
    no atoms present, just prevents conflicts of obj.key; atom present, assumes
    it's a blocker and ensures no obj matches the atom for that key
    """

    def __init__(self):
        self.slot_dict = {}
        self.limiters = {}

    def fill_slotting(self, obj, force=False):
        """Try to insert obj in.

        :return: any conflicting objs (empty list if inserted successfully).
        """

        l = self.check_limiters(obj)

        key = obj.key
        dslot = obj.slot
        l.extend(x for x in self.slot_dict.get(key, ()) if x.slot == dslot)

        if not l or force:
            self.slot_dict.setdefault(key, []).append(obj)
        return l

    def get_conflicting_slot(self, pkg):
        for x in self.slot_dict.get(pkg.key, ()):
            if pkg.slot == x.slot:
                return x
        return None

    def find_atom_matches(self, atom, key=None):
        if key is None:
            key = atom.key
        return list(filter(atom.match, self.slot_dict.get(key, ())))

    def add_limiter(self, atom, key=None):
        """add a limiter, returning any conflicting objs"""
        if not isinstance(atom, restriction.base):
            raise TypeError(
                f"atom must be a restriction.base derivative: got {atom!r}, key={key!r}")
        # debug.

        if key is None:
            key = atom.key
        self.limiters.setdefault(key, []).append(atom)
        return self.find_atom_matches(atom, key=key)

    def check_limiters(self, obj):
        """return any limiters conflicting w/ the passed in obj"""
        key = obj.key
        return [x for x in self.limiters.get(key, ()) if x.match(obj)]

    def remove_slotting(self, obj):
        key = obj.key
        # let the key error be thrown if they screwed up.
        slots = self.slot_dict.get(key, ())
        l = [x for x in slots if x is not obj]
        if len(l) == len(slots):
            raise KeyError(f"obj {obj} isn't slotted")
        if l:
            self.slot_dict[key] = l
        else:
            del self.slot_dict[key]

    def remove_limiter(self, atom, key=None):
        if key is None:
            key = atom.key
        l = [x for x in self.limiters[key] if x is not atom]
        if len(l) == len(self.limiters[key]):
            raise KeyError(f"obj {atom} isn't slotted")
        if not l:
            del self.limiters[key]
        else:
            self.limiters[key] = l

    def __contains__(self, obj):
        if isinstance(obj, restriction.base):
            return obj in self.limiters.get(obj.key, ())
        return obj in self.slot_dict.get(obj.key, ())



from pkgcore.ebuild import eclass
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import TempDirMixin


class FakeEclass:
    def __init__(self, path, contents):
        self.path = path
        with open(path, 'w') as f:
            f.write(contents)


class FakeEclassCache:
    def __init__(self, temp_dir, eclasses):
        self.eclasses = dict((name, FakeEclass(name, contents))
                             for name, contents in eclasses.items())

    def get_eclass(self, name):
        return self.eclasses.get(name)


class FakeEclassRepo:
    def __init__(self, temp_dir, eclasses):
        self.eclass_cache = FakeEclassCache(temp_dir, eclasses)


def make_eclass(name, provides=None):
    eclass = f'# @ECLASS: {name}.eclass\n'
    if provides is not None:
        eclass += f'# @PROVIDES: {provides}\n'
    return eclass


FOO_ECLASS = '''
# @ECLASS: foo.eclass
# @MAINTAINER:
# Random Person <maintainer@random.email>
# @AUTHOR:
# Another Person <another@random.email>
# Random Person <maintainer@random.email>
# @BUGREPORTS:
# Report bugs somewhere.
# @VCSURL: https://example.com/foo.eclass
# @SUPPORTED_EAPIS: 0 1 2 3 4 5 6 7
# @PROVIDES: bar
# @BLURB: Test eclass.
# @DEPRECATED: bar or frobnicate
# @DESCRIPTION:
# Yadda yadda yadda.
# Lots to say here.
#
# Really, very interesting eclass.
#
# @SUBSECTION How to use it
# Somehow.
#
# @EXAMPLE:
# @CODE
# inherit foo
#
# src_prepare() {
#   foo_public_func
# }
# @CODE

# @ECLASS_VARIABLE: _FOO_INTERNAL_ECLASS_VAR
# @INTERNAL
# @DEFAULT_UNSET
# @DESCRIPTION:
# Internal variable.

# @ECLASS_VARIABLE: FOO_PUBLIC_ECLASS_VAR
# @PRE_INHERIT
# @REQUIRED
# @DEPRECATED: BAR_PUBLIC_ECLASS_VAR
# @DESCRIPTION:
# Public variable.

# @ECLASS-VARIABLE: FOO_ANOTHER_ECLASS_VAR
# @DESCRIPTION:
# Yet another variable.

# @FUNCTION: _foo_internal_func
# @USAGE: <bar> [<baz>]
# @RETURN: nothing special
# @MAINTAINER:
# Some Person <someone@random.email>
# @INTERNAL
# @DESCRIPTION:
# Internal stub function.
_foo_internal_func() { :; }

# @VARIABLE: _FOO_INTERNAL_VAR
# @INTERNAL
# @DEFAULT_UNSET
# @DESCRIPTION:
# Internal variable for foo_public_func.

# @VARIABLE: FOO_PUBLIC_VAR
# @REQUIRED
# @DEPRECATED: BAR_PUBLIC_VAR
# @DESCRIPTION:
# Public variable for foo_public_func.

# @FUNCTION: foo_public_func
# @DEPRECATED: bar_public_func
# @DESCRIPTION:
# Public stub function.
foo_public_func() { :; }
'''


class TestEclassDoc(TempDirMixin, TestCase):
    def test_foo_eclass(self):
        with open(pjoin(self.dir, 'foo.eclass'), 'w') as f:
            f.write(FOO_ECLASS)
        doc = eclass.EclassDoc(pjoin(self.dir, 'foo.eclass'))
        assert doc.name == 'foo.eclass'
        assert doc.vcsurl == 'https://example.com/foo.eclass'
        assert doc.blurb == 'Test eclass.'
        assert doc.deprecated == 'bar or frobnicate'
        assert doc.raw_provides == ('bar',)
        assert doc.maintainers == ('Random Person <maintainer@random.email>',)
        assert doc.authors == ('Another Person <another@random.email>',
                               'Random Person <maintainer@random.email>',)
        assert doc.bugreports == '::\n\n  Report bugs somewhere.'
        assert doc.description == ('::\n\n'
                                   '  Yadda yadda yadda.\n'
                                   '  Lots to say here.\n\n'
                                   '  Really, very interesting eclass.\n\n'
                                   'How to use it\n'
                                   '~~~~~~~~~~~~~\n'
                                   '::\n\n'
                                   '  Somehow.')
        assert doc.example == ('::\n\n'
                               '  inherit foo\n\n'
                               '  src_prepare() {\n'
                               '    foo_public_func\n'
                               '  }')
        assert doc.supported_eapis == frozenset(map(str, range(8)))

        assert doc.function_names == frozenset(('_foo_internal_func', 'foo_public_func'))
        assert doc.internal_function_names == frozenset(('_foo_internal_func',))

        assert doc.function_variable_names == frozenset(('FOO_PUBLIC_VAR', '_FOO_INTERNAL_VAR'))

        assert doc.variable_names == frozenset(('FOO_PUBLIC_ECLASS_VAR',
                                                '_FOO_INTERNAL_ECLASS_VAR',
                                                'FOO_ANOTHER_ECLASS_VAR'))
        assert doc.internal_variable_names == frozenset(('_FOO_INTERNAL_ECLASS_VAR',))

        assert len(doc.functions) == 2
        assert doc.functions[0] == {'name': '_foo_internal_func',
                                    'returns': 'nothing special',
                                    'deprecated': False,
                                    'internal': True,
                                    'maintainers': ('Some Person <someone@random.email>',),
                                    'description': '::\n\n  Internal stub function.',
                                    'usage': '<bar> [<baz>]'}

        assert doc.functions[1] == {'name': 'foo_public_func',
                                    'returns': None,
                                    'deprecated': 'bar_public_func',
                                    'internal': False,
                                    'maintainers': None,
                                    'description': '::\n\n  Public stub function.',
                                    'usage': None}

        assert len(doc.function_variables) == 2
        assert doc.function_variables[0] == {'name': '_FOO_INTERNAL_VAR',
                                             'deprecated': False,
                                             'default_unset': True,
                                             'internal': True,
                                             'required': False,
                                             'description': '::\n\n  Internal variable for foo_public_func.'}
        assert doc.function_variables[1] == {'name': 'FOO_PUBLIC_VAR',
                                             'deprecated': 'BAR_PUBLIC_VAR',
                                             'default_unset': False,
                                             'internal': False,
                                             'required': True,
                                             'description': '::\n\n  Public variable for foo_public_func.'}

        assert len(doc.variables) == 3
        assert doc.variables[0] == {'name': '_FOO_INTERNAL_ECLASS_VAR',
                                    'deprecated': False,
                                    'default_unset': True,
                                    'internal': True,
                                    'required': False,
                                    'pre_inherit': False,
                                    'user_variable': False,
                                    'output_variable': False,
                                    'description': '::\n\n  Internal variable.'}
        assert doc.variables[1] == {'name': 'FOO_PUBLIC_ECLASS_VAR',
                                    'deprecated': 'BAR_PUBLIC_ECLASS_VAR',
                                    'default_unset': False,
                                    'internal': False,
                                    'required': True,
                                    'pre_inherit': True,
                                    'user_variable': False,
                                    'output_variable': False,
                                    'description': '::\n\n  Public variable.'}
        assert doc.variables[2] == {'name': 'FOO_ANOTHER_ECLASS_VAR',
                                    'deprecated': False,
                                    'default_unset': False,
                                    'internal': False,
                                    'required': False,
                                    'pre_inherit': False,
                                    'user_variable': False,
                                    'output_variable': False,
                                    'description': '::\n\n  Yet another variable.'}

    def test_recursive_provides(self):
        repo = FakeEclassRepo(self.dir, {
            'foo': FOO_ECLASS,
            'bar': make_eclass('bar', provides='deep1 deep2'),
            'deep1': make_eclass('deep1 deep2'),
            'deep2': make_eclass('deep2 foo'),
        })
        assert (sorted(eclass.EclassDoc(repo.eclass_cache.get_eclass('foo').path,
                                        repo=repo).provides) ==
                ['bar', 'deep1', 'deep2'])



import os
import textwrap
from functools import partial

import pytest
from pkgcore import fetch
from pkgcore.ebuild import digest, ebuild_src, repo_objs
from pkgcore.ebuild.eapi import EAPI, get_eapi
from pkgcore.package import errors
from pkgcore.test import malleable_obj
from snakeoil.currying import post_curry
from snakeoil.data_source import data_source, local_source
from snakeoil.osutils import pjoin

from .test_eclass_cache import FakeEclassCache


class TestBase:

    kls = ebuild_src.base

    def get_pkg(self, data=None, cpv='dev-util/diffball-0.1-r1', repo=None,
                pre_args=(), suppress_unsupported=True):
        o = self.kls(*(list(pre_args) + [repo, cpv]))
        if data is not None:
            eapi_data = data.pop('EAPI', 0)
            if eapi_data is not None:
                object.__setattr__(o, 'eapi', get_eapi(
                    str(eapi_data), suppress_unsupported=suppress_unsupported))
            object.__setattr__(o, 'data', data)
        return o

    def make_parent(self, **methods):
        class kls:
            locals().update(methods)
        return kls()

    def test_init(self):
        o = self.get_pkg({}, cpv='dev-util/diffball-0.1-r1')
        assert o.category == 'dev-util'
        assert o.package == 'diffball'
        assert o.fullver == '0.1-r1'
        assert o.PN == 'diffball'
        assert o.P == 'diffball-0.1'
        assert o.PF == 'diffball-0.1-r1'
        assert o.PR == 'r1'
        assert self.get_pkg({}, 'dev-util/diffball-0.1').PR == 'r0'

    def test_path(self):
        l = []
        path = '/random/path/to/foo-0.ebuild'
        def f(self, cpv):
            l.append(cpv)
            return path
        c = self.make_parent(_get_ebuild_path=f)
        o = self.get_pkg({}, repo=c)
        assert o.path == path
        assert l == [o]

    def test_ebuild(self):
        l = []
        def f(self, cpv):
            l.append(cpv)
            return 1
        c = self.make_parent(get_ebuild_src=f)
        o = self.get_pkg({}, repo=c)
        assert o.ebuild == 1
        assert l == [o]

    def test_fetch_metadata(self):
        def f(self, cpv, **options):
            return {'1': '2'}
        o = self.get_pkg(repo=self.make_parent(_get_metadata=f))
        assert o.data == {'1': '2'}

    def test_license(self):
        o = self.get_pkg({'LICENSE': 'GPL2 FOON'})
        assert list(o.license) == ['GPL2', 'FOON']

    def test_description(self):
        o = self.get_pkg({'DESCRIPTION': ' foon\n asdf '})
        assert o.description == 'foon\n asdf'

    def test_iuse(self):
        o = self.get_pkg({})
        assert o.iuse == frozenset()
        o = self.get_pkg({'IUSE': 'build pkg foon'})
        assert o.iuse == frozenset(['build', 'foon', 'pkg'])

    def test_iuse_stripped(self):
        o = self.get_pkg({})
        assert o.iuse_stripped == frozenset()
        o = self.get_pkg({'IUSE': 'build pkg foon'})
        assert o.iuse_stripped == frozenset(['build', 'foon', 'pkg'])
        o = self.get_pkg({'EAPI': '1', 'IUSE': '+build -pkg foon'})
        assert o.iuse_stripped == frozenset(['build', 'foon', 'pkg'])

    def test_iuse_effective(self):
        o = self.get_pkg({})
        assert o.iuse_effective == frozenset()
        o = self.get_pkg({'IUSE': 'build pkg foon'})
        assert o.iuse_effective == frozenset(['build', 'foon', 'pkg'])
        o = self.get_pkg({'EAPI': '1', 'IUSE': '+build -pkg foon'})
        assert o.iuse_effective == frozenset(['build', 'foon', 'pkg'])

    def test_properties(self):
        o = self.get_pkg({})
        assert sorted(o.properties.evaluate_depset([])) == []
        o = self.get_pkg({'PROPERTIES': ''})
        assert sorted(o.properties.evaluate_depset([])) == []
        o = self.get_pkg({'PROPERTIES': 'interactive'})
        assert sorted(o.properties.evaluate_depset([])) == ['interactive']

    def test_homepage(self):
        o = self.get_pkg({'HOMEPAGE': ' http://slashdot/ '})
        assert o.homepage == ('http://slashdot/',)
        o = self.get_pkg({'HOMEPAGE': 'http://foozball.org https://foobar.com'})
        assert o.homepage == ('http://foozball.org', 'https://foobar.com')

    def test_fullslot(self):
        o = self.get_pkg({'SLOT': '0'})
        assert o.fullslot == '0'

        # subslot support
        for eapi_str, eapi in EAPI.known_eapis.items():
            if eapi.options.sub_slotting:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '0/0'})
                assert o.fullslot == '0/0'
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/2'})
                assert o.fullslot == '1/2'
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/foo-1'})
                assert o.fullslot == '1/foo-1'
            else:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '0/0'})
                with pytest.raises(errors.MetadataException):
                    o.fullslot

        # unset SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({}).fullslot
        # empty SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({'SLOT': ''}).fullslot

    def test_slot(self):
        o = self.get_pkg({'SLOT': '0'})
        assert o.slot == '0'

        # subslot support
        for eapi_str, eapi in EAPI.known_eapis.items():
            if eapi.options.sub_slotting:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/2'})
                assert o.slot == '1'
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/foo-1'})
                assert o.slot == '1'
            else:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/2'})
                with pytest.raises(errors.MetadataException):
                    o.slot

        # unset SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({}).slot
        # empty SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({'SLOT': ''}).slot

    def test_subslot(self):
        o = self.get_pkg({'SLOT': '0'})
        assert o.subslot == '0'
        o = self.get_pkg({'SLOT': '1'})
        assert o.subslot == '1'

        # subslot support
        for eapi_str, eapi in EAPI.known_eapis.items():
            if eapi.options.sub_slotting:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/2'})
                assert o.subslot == '2'
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/foo-1'})
                assert o.subslot == 'foo-1'
            else:
                o = self.get_pkg({'EAPI': eapi_str, 'SLOT': '1/2'})
                with pytest.raises(errors.MetadataException):
                    o.subslot

        # unset SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({}).subslot
        # empty SLOT variable
        with pytest.raises(errors.MetadataException):
            self.get_pkg({'SLOT': ''}).subslot

    def test_restrict(self):
        o = self.get_pkg({'RESTRICT': 'strip fetch strip'})
        assert sorted(o.restrict) == ['fetch', 'strip', 'strip']
        o = self.get_pkg({'RESTRICT': 'x? ( foo ) !x? ( dar )'})
        assert sorted(o.restrict.evaluate_depset([])) == ['dar']
        # ensure restrict doesn't have || () in it
        with pytest.raises(errors.MetadataException):
            getattr(self.get_pkg({'RESTRICT': '|| ( foon dar )'}), 'restrict')

    def test_eapi(self):
        assert str(self.get_pkg({'EAPI': '0'}).eapi) == '0'
        assert self.get_pkg({'EAPI': '0'}).eapi.is_supported
        assert not self.get_pkg({'EAPI': '0.1'}).eapi.is_supported
        assert self.get_pkg({'EAPI': 'foon'}, suppress_unsupported=False).eapi is None
        with pytest.raises(errors.MetadataException):
            getattr(self.get_pkg({'EAPI': 0, 'DEPEND': "d/b:0"}), 'depend')
        with pytest.raises(errors.MetadataException):
            getattr(self.get_pkg({'EAPI': 0, 'RDEPEND': "d/b:0"}), 'rdepend')
        with pytest.raises(errors.MetadataException):
            getattr(self.get_pkg({'EAPI': 1, 'DEPEND': "d/b[x,y]"}), 'depend')
        with pytest.raises(errors.MetadataException):
            getattr(self.get_pkg({'EAPI': 1, 'DEPEND': "d/b::foon"}), 'depend')
        assert self.get_pkg({'EAPI': 2, 'DEPEND': 'a/b[x=]'}).depend.node_conds
        pkg = self.get_pkg({'EAPI': 1, 'DEPEND': 'a/b[x=]'})
        with pytest.raises(errors.MetadataException):
            getattr(pkg, 'depend')

    def test_get_parsed_eapi(self, tmpdir):
        # ebuild has a real path on the fs
        def _path(self, cpv, eapi_str):
            ebuild = pjoin(str(tmpdir), "temp-0.ebuild")
            with open(ebuild, 'w') as f:
                f.write(textwrap.dedent(f'''\
                    # Copyright
                    # License

                    EAPI={eapi_str}'''))
            return local_source(str(ebuild))

        # ebuild is a faked obj
        def _src(self, cpv, eapi_str):
            return data_source(f'EAPI={eapi_str}')

        for func in (_path, _src):
            # verify parsing known EAPIs
            for eapi_str in EAPI.known_eapis.keys():
                c = self.make_parent(get_ebuild_src=post_curry(func, eapi_str))
                o = self.get_pkg({'EAPI': None}, repo=c)
                assert str(o.eapi) == eapi_str
            # check explicitly unsetting EAPI equates to EAPI=0
            for eapi_str in ('', '""', "''"):
                c = self.make_parent(get_ebuild_src=post_curry(func, eapi_str))
                o = self.get_pkg({'EAPI': None}, repo=c)
                assert str(o.eapi) == '0'

    def test_keywords(self):
        assert list(self.get_pkg({'KEYWORDS': ''}).keywords) == []
        assert sorted(self.get_pkg({'KEYWORDS': 'x86 amd64'}).keywords) == sorted(['x86', 'amd64'])

    def test_sorted_keywords(self):
        assert self.get_pkg({'KEYWORDS': ''}).sorted_keywords == ()
        assert self.get_pkg({'KEYWORDS': 'amd64 x86'}).sorted_keywords == ('amd64', 'x86')
        assert self.get_pkg({'KEYWORDS': 'x86 amd64'}).sorted_keywords == ('amd64', 'x86')
        assert (
            self.get_pkg({'KEYWORDS': '~amd64 ~amd64-fbsd ~x86'}).sorted_keywords ==
            ('~amd64', '~x86', '~amd64-fbsd'))
        assert (
            self.get_pkg({'KEYWORDS': '~amd64 ~x86 ~amd64-fbsd'}).sorted_keywords ==
            ('~amd64', '~x86', '~amd64-fbsd'))

    def generic_check_depends(self, depset, attr, expected=None,
                              data_name=None, eapi='0'):
        if expected is None:
            expected = depset
        if data_name is None:
            data_name = attr.upper()
        o = self.get_pkg({data_name: depset, 'EAPI': eapi})
        assert str(getattr(o, attr)) == expected
        o = self.get_pkg({data_name: '', 'EAPI': eapi})
        assert str(getattr(o, attr)) == ''
        if expected:
            with pytest.raises(errors.MetadataException):
                getattr(self.get_pkg({data_name: '|| ( ', 'EAPI': eapi}), attr)

    for x in ('depend', 'rdepend'):
        locals()[f'test_{x}'] = post_curry(generic_check_depends,
            'dev-util/diffball || ( dev-util/foo x86? ( dev-util/bsdiff ) )',
             x)
    del x
    test_pdepend = post_curry(generic_check_depends,
        'dev-util/diffball x86? ( virtual/boo )', 'pdepend')
    # BDEPEND in EAPI 7
    test_bdepend = post_curry(generic_check_depends,
        'dev-util/diffball x86? ( virtual/boo )', 'bdepend', eapi='7')
    # BDEPEND is ignored in EAPIs <= 6
    test_no_bdepend = post_curry(generic_check_depends,
        'dev-util/diffball x86? ( virtual/boo )', 'bdepend', expected='', eapi='0')
    # IDEPEND in EAPI 8
    test_idepend = post_curry(generic_check_depends,
        'dev-util/diffball x86? ( virtual/boo )', 'idepend', eapi='8')
    # IDEPEND is ignored in EAPIs <= 7
    test_no_idepend = post_curry(generic_check_depends,
        'dev-util/diffball x86? ( virtual/boo )', 'idepend', expected='', eapi='0')

    def test_fetchables(self):
        l = []
        def f(self, cpv, allow_missing=False):
            l.append(cpv)
            return allow_missing, {'monkey.tgz': {}, 'boon.tgz': {}, 'foon.tar.gz': {}}
        repo = self.make_parent(_get_digests=f)
        parent = self.make_parent(_parent_repo=repo)
        # verify it does digest lookups...
        o = self.get_pkg({'SRC_URI': 'http://foo.com/bar.tgz'}, repo=parent)
        with pytest.raises(errors.MetadataException):
            getattr(o, 'fetchables')
        assert l == [o]

        # basic tests;
        for x in range(0, 3):
            f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz',
                'EAPI': str(x)},
                 repo=parent).fetchables
            assert list(f[0].uri) == ['http://foo.com/monkey.tgz']
            assert f[0].filename == 'monkey.tgz'

        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz '
            'http://dar/boon.tgz', 'EAPI': '2'},
             repo=parent).fetchables
        assert [list(x.uri) for x in f] == [['http://foo.com/monkey.tgz'], ['http://dar/boon.tgz']]
        assert [x.filename for x in f] == ['monkey.tgz', 'boon.tgz']

        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz -> foon.tar.gz',
            'EAPI': '2'},
             repo=parent).fetchables
        assert list(f[0].uri) == ['http://foo.com/monkey.tgz']
        assert f[0].filename == 'foon.tar.gz'

        o = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz -> ',
            'EAPI': '2'}, repo=parent)
        with pytest.raises(errors.MetadataException):
            getattr(o, 'fetchables')

        # verify it collapses multiple basenames down to the same.
        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz '
            'http://foo.com2/monkey.tgz'}, repo=parent).fetchables
        assert list(f[0].uri) == ['http://foo.com/monkey.tgz', 'http://foo.com2/monkey.tgz']

        mirror = fetch.mirror(['http://boon.com/'], 'mirror1')
        parent = self.make_parent(_parent_repo=repo, mirrors={'mirror1': mirror})

        f = self.get_pkg(
            {'SRC_URI': 'mirror://mirror1/foon/monkey.tgz'}, repo=parent).fetchables
        assert list(f[0].uri) == ['http://boon.com/foon/monkey.tgz']

        # unknown mirrors get ignored (and flagged by pkgcheck)
        pkg = self.get_pkg({'SRC_URI': 'mirror://mirror2/foon/monkey.tgz'}, repo=parent)
        assert pkg.fetchables

        assert (
            [list(x.uri) for x in self.get_pkg(
                {'EAPI': '2', 'SRC_URI': 'mirror://mirror1/monkey.tgz -> foon.tar.gz'},
                repo=parent).fetchables] ==
            [['http://boon.com/monkey.tgz']])

        parent = self.make_parent(_parent_repo=repo,
            mirrors={'mirror1': mirror}, default_mirrors=fetch.default_mirror(
                ['http://default.com/dist/', 'http://default2.com/'],
                'default'))

        assert (
            [list(x.uri) for x in self.get_pkg(
                {'EAPI': '2', 'SRC_URI': 'mirror://mirror1/monkey.tgz -> foon.tar.gz'},
                repo=parent).fetchables] ==
            [[
                'http://default.com/dist/foon.tar.gz',
                'http://default2.com/foon.tar.gz',
                'http://boon.com/monkey.tgz']])

        parent = self.make_parent(_parent_repo=repo, default_mirrors=mirror)
        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz'},
            repo=parent).fetchables
        assert list(f[0].uri) == ['http://boon.com/monkey.tgz', 'http://foo.com/monkey.tgz']

        # skip default mirrors
        pkg = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz'}, repo=parent)
        f = pkg.generate_fetchables(skip_default_mirrors=True)
        assert list(f[0].uri) == ['http://foo.com/monkey.tgz']

        # test primaryuri...
        mirror2 = fetch.mirror(['http://boon2.com/'], 'default')
        parent = self.make_parent(_parent_repo=repo, default_mirrors=mirror,
            mirrors={'mirror1': mirror2})
        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz '
            'mirror://mirror1/boon.tgz', 'RESTRICT': 'primaryuri'},
            repo=parent).fetchables
        assert list(f[0].uri) == ['http://foo.com/monkey.tgz', 'http://boon.com/monkey.tgz']
        assert list(f[1].uri) == ['http://boon2.com/boon.tgz', 'http://boon.com/boon.tgz']
        assert len(f) == 2

        # restrict=mirror..
        f = self.get_pkg({'SRC_URI': 'http://foo.com/monkey.tgz',
            'RESTRICT': 'mirror'}, repo=parent).fetchables
        assert list(f[0].uri) == ['http://foo.com/monkey.tgz']
        assert len(f) == 1

        # test uri for when there *is* no uri
        f = self.get_pkg({'SRC_URI': 'monkey.tgz'}, repo=parent).fetchables
        assert list(f[0].uri) == []

    def test_required_use(self):
        for eapi_str, eapi in EAPI.known_eapis.items():
            # Check all EAPIs for REQUIRED_USE parsing, EAPIs that don't support it
            # should return depsets that evaluate to False.
            pkg = self.get_pkg({'EAPI': eapi_str, 'REQUIRED_USE': 'test? ( foo )'})
            assert bool(pkg.required_use) == eapi.options.has_required_use, \
                f"failure parsing REQUIRED_USE for EAPI '{eapi}'"

            # Render various REQUIRED_USE deps with set USE flag states and
            # check for satisfiability.
            if eapi.options.has_required_use:
                required_use_data = (
                    ('foo', '', 'foo', True),
                    ('foo', '', '', False),
                    ('!foo', 'foo', '', True),
                    ('!foo', 'foo', 'foo', False),
                    ('foo bar', '', 'foo bar', True),
                    ('foo bar', '', 'bar foo', True),
                    ('( foo bar )', '', 'foo', False),
                    ('( foo bar )', '', 'foo bar', True),
                    ('test? ( foo )', 'test', 'foo', True),
                    ('test? ( foo )', 'test', '', False),
                    ('test? ( !foo )', 'test', 'foo', False),
                    ('test? ( !foo )', 'test', '', True),
                    ('test? ( foo bar )', 'test', 'foo bar', True),
                    ('!test? ( foo )', 'test', '', True),
                    ('!test? ( foo )', '', 'foo', True),
                    ('|| ( test foo )', 'test', 'test', True),
                    ('|| ( test foo )', 'test', 'test foo', True),
                    ('|| ( test foo ) bar? ( foo )', 'test', 'test', True),
                    ('|| ( test foo ) bar? ( foo )', 'bar', 'foo', True),
                    ('^^ ( bar foo )', '', 'bar', True),
                    ('^^ ( bar foo )', '', 'foo', True),
                    ('^^ ( bar foo )', '', '', False),
                )

                if eapi.options.required_use_one_of:
                    required_use_data += (
                        ('?? ( bar foo )', '', 'bar', True),
                        ('?? ( bar foo )', '', 'foo', True),
                        ('?? ( bar foo )', '', '', True),
                    )
                else:
                    # EAPIs that don't support the ?? operator raise metadata exceptions if used.
                    pkg = self.get_pkg({'EAPI': eapi_str, 'REQUIRED_USE': '?? ( bar foo )'})
                    with pytest.raises(errors.MetadataException) as cm:
                        getattr(pkg, 'required_use')
                    assert f"EAPI '{eapi_str}' doesn't support '??' operator" in cm.value.error

                for required_use, iuse, use, satisfied in required_use_data:
                    pkg = self.get_pkg({'EAPI': eapi_str, 'REQUIRED_USE': required_use})
                    required_use_deps = pkg.required_use.evaluate_depset(iuse.split())
                    for node in required_use_deps:
                        assert node.match(use.split()) is satisfied, \
                            f'REQUIRED_USE="{required_use}", IUSE="{iuse}", ' \
                            f'USE="{use}", satisfied="{satisfied}"'

    def test_live(self):
        o = self.get_pkg({})
        assert not o.live
        o = self.get_pkg({'PROPERTIES': 'live'})
        assert o.live


class TestPackage(TestBase):

    kls = ebuild_src.package

    def get_pkg(self, *args, **kwds):
        kwds.setdefault("pre_args", (None,))
        return super().get_pkg(*args, **kwds)

    def test_init(self):
        super().test_init()
        o = self.get_pkg(pre_args=(1,))
        assert o._shared_pkg_data == 1

    def test_mtime_(self):
        l = []
        def f(self, cpv):
            l.append(cpv)
            return 100

        parent = self.make_parent(_get_ebuild_mtime=f)
        o = self.get_pkg(repo=parent)
        assert o._mtime_ == 100
        assert l == [o]

    def make_shared_pkg_data(self, manifest=None, metadata_xml=None):
        return self.get_pkg(
            pre_args=(repo_objs.SharedPkgData(metadata_xml, manifest),))

    def generic_metadata_xml(self, attr):
        m = repo_objs.MetadataXml(None)
        object.__setattr__(m, "_"+attr, "foon")
        object.__setattr__(m, "_source", None)
        o = self.make_shared_pkg_data(metadata_xml=m)
        assert getattr(o, attr) == "foon"

    for x in ("longdescription", "maintainers"):
        locals()[f"test_{x}"] = post_curry(generic_metadata_xml, x)
    del x

    def test_manifest(self):
        m = digest.Manifest(None)
        o = self.make_shared_pkg_data(manifest=m)
        assert o.manifest is m


class TestPackageFactory:

    kls = ebuild_src.package_factory

    def mkinst(self, repo=None, cache=(), eclasses=None, mirrors={},
               default_mirrors={}, **overrides):
        o = self.kls(repo, cache, eclasses, mirrors, default_mirrors)
        for k, v in overrides.items():
            object.__setattr__(o, k, v)
        return o

    def test_mirrors(self):
        mirrors_d = {'gentoo': ['http://bar/', 'http://far/']}
        mirrors = {k: fetch.mirror(v, k) for k, v in mirrors_d.items()}
        pf = self.mkinst(mirrors=mirrors_d)
        assert len(pf._cache) == 0
        assert sorted(pf.mirrors) == sorted(mirrors)
        assert pf.mirrors['gentoo'] == mirrors['gentoo']
        assert pf.default_mirrors == None

        def_mirrors = ['http://def1/', 'http://blah1/']
        pf = self.mkinst(default_mirrors=def_mirrors)
        assert pf.mirrors == {}
        assert list(pf.default_mirrors) == def_mirrors

    def test_get_ebuild_src(self):
        assert (
            self.mkinst(
                repo=malleable_obj(_get_ebuild_src=lambda s: f"lincoln haunts me: {s}")
                ).get_ebuild_src("1") ==
            "lincoln haunts me: 1")

    def test_get_ebuild_mtime(self, tmpdir):
        f = pjoin(str(tmpdir), "temp-0.ebuild")
        open(f, 'w').close()
        mtime = self.mkinst(
            repo=malleable_obj(_get_ebuild_path=lambda s: f))._get_ebuild_mtime(None)
        assert mtime == os.stat(f).st_mtime

    def test_get_metadata(self):
        ec = FakeEclassCache('/nonexistent/path')
        pkg = malleable_obj(_mtime_=100, cpvstr='dev-util/diffball-0.71', path='bollocks')

        class fake_cache(dict):
            readonly = False
            validate_result = False
            def validate_entry(self, *args):
                return self.validate_result

        cache1 = fake_cache({pkg.cpvstr: {'_mtime_': 100, 'marker': 1}})
        cache2 = fake_cache({})

        class explode_kls(AssertionError): pass

        def explode(name, *args, **kwargs):
            raise explode_kls(
                f"{name} was called with {args!r} and {kwargs!r}, shouldn't be invoked.")

        pf = self.mkinst(
            cache=(cache2, cache1), eclasses=ec,
            _update_metadata=partial(explode, '_update_metadata'))

        cache1.validate_result = True
        assert pf._get_metadata(pkg) == {'marker': 1, '_mtime_': 100}

        assert list(cache1.keys()) == [pkg.cpvstr]
        assert not cache2

        # mtime was wiped, thus no longer is usable.
        # note also, that the caches are writable.
        cache1.validate_result = False
        with pytest.raises(explode_kls):
            pf._get_metadata(pkg)
        assert not cache2
        assert not cache1

        # Note that this is known crap eclass data; partially lazyness, partially
        # to validate the eclass validation is left to ec cache only.
        cache2.update({pkg.cpvstr:
            {'_mtime_': 200, '_eclasses_': {'eclass1': (None, 100)}, 'marker': 2}
        })
        cache2.readonly = True
        with pytest.raises(explode_kls):
            pf._get_metadata(pkg)
        assert list(cache2.keys()) == [pkg.cpvstr]
        # keep in mind the backend assumes it gets its own copy of the data.
        # thus, modifying (popping _mtime_) _is_ valid
        assert cache2[pkg.cpvstr] == \
            {'_eclasses_': {'eclass1': (None, 100)}, 'marker': 2, '_mtime_': 200}



__all__ = ("resolver_frame", "resolver_stack", "merge_plan")

import operator
import sys
from collections import deque
from functools import partial
from itertools import chain, filterfalse, islice

from snakeoil.compatibility import cmp, sort_cmp
from snakeoil.iterables import caching_iter

# XXX: hack; see insert_blockers
from ..ebuild import atom as _atom
from ..repository import filtered, misc, multiplex, util
from ..restrictions import packages, restriction, values
from . import state
from .choice_point import choice_point

limiters = set(["cycle"])


def dprint(handle, fmt, args=None, label=None):
    if None in limiters or label in limiters:
        if args:
            fmt = fmt % args
        handle.write(fmt)
        handle.write("\n")


# iter/pkg sorting functions for selection strategy
pkg_sort_highest = partial(sorted, reverse=True)
pkg_sort_lowest = sorted

pkg_grabber = operator.itemgetter(0)


def highest_iter_sort(l, pkg_grabber=pkg_grabber):
    """Sort a list of packages from highest to lowest and prefer livefs.

    :param l: list of packages
    :param pkg_grabber: function to use as an attrgetter
    :return: sorted list of packages
    """
    def f(x, y):
        c = cmp(x, y)
        if c:
            return c
        elif x.repo.livefs:
            if y.repo.livefs:
                return 0
            return 1
        elif y.repo.livefs:
            return -1
        return 0
    sort_cmp(l, f, key=pkg_grabber, reverse=True)
    return l


def downgrade_iter_sort(restrict, l, pkg_grabber=pkg_grabber):
    """Sort a list of packages from highest to lowest and prefer nonlivefs.

    :param l: list of packages
    :param pkg_grabber: function to use as an attrgetter
    :return: sorted list of packages
    """
    def f(x, y):
        c = cmp(x, y)
        if x.repo.livefs:
            if y.repo.livefs:
                return c
            return -1
        elif y.repo.livefs:
            return 1
        elif restrict.match(x):
            if restrict.match(y):
                return 1
            return -1
        elif restrict.match(y):
            return 1
        return c
    sort_cmp(l, f, key=pkg_grabber, reverse=True)
    return l


def lowest_iter_sort(l, pkg_grabber=pkg_grabber):
    """Sort a list of packages from lowest to highest.

    :param l: list of packages
    :param pkg_grabber: function to use as an attrgetter
    :return: sorted list of packages
    """
    def f(x, y):
        c = cmp(x, y)
        if c:
            return c
        elif x.repo.livefs:
            if y.repo.livefs:
                return 0
            return -1
        elif y.repo.livefs:
            return 1
        return 0
    sort_cmp(l, f, key=pkg_grabber)
    return l


class MutableContainmentRestriction(values.base):

    __slots__ = ('_blacklist', 'match')

    def __init__(self, blacklist):
        sf = object.__setattr__
        sf(self, '_blacklist', blacklist)
        sf(self, 'match', self._blacklist.__contains__)


class resolver_frame:

    __slots__ = ("parent", "atom", "choices", "mode", "start_point", "dbs",
        "depth", "drop_cycles", "__weakref__", "ignored", "vdb_limited",
        "events", "succeeded")

    def __init__(self, parent, mode, atom, choices, dbs, start_point, depth,
                 drop_cycles, ignored=False, vdb_limited=False):
        assert hasattr(dbs, 'itermatch')
        self.parent = parent
        self.atom = atom
        self.choices = choices
        self.dbs = dbs
        self.mode = mode
        self.start_point = start_point
        self.depth = depth
        self.drop_cycles = drop_cycles
        self.ignored = False
        self.vdb_limited = vdb_limited
        self.events = []
        self.succeeded = None

    def reduce_solutions(self, nodes):
        if isinstance(nodes, (list, tuple)):
            self.events.append(("reduce", nodes))
        else:
            self.events.append(("reduce", (nodes,)))
        return self.choices.reduce_atoms(nodes)

    def __str__(self):
        pkg = self.current_pkg
        if pkg is None:
            pkg = "exhausted"
        else:
            cpv = pkg.cpvstr
            pkg = getattr(pkg.repo, 'repo_id', None)
            if pkg is not None:
                pkg = f"{cpv}::{pkg}"
            else:
                pkg = str(pkg)
        if self.succeeded is not None:
            result = ": %s" % (self.succeeded and "succeeded" or "failed")
        else:
            result = ""
        return "frame%s: mode %r: atom %s: current %s%s%s%s" % \
            (result, self.mode, self.atom, pkg,
            self.drop_cycles and ": cycle dropping" or '',
            self.ignored and ": ignored" or '',
            self.vdb_limited and ": vdb limited" or '')

    @property
    def current_pkg(self):
        """Return the package related to the resolver frame."""
        try:
            return self.choices.current_pkg
        except IndexError:
            return None


class resolver_stack(deque):

    frame_klass = resolver_frame
    depth = property(len)
    current_frame = property(operator.itemgetter(-1))
    _filter_ignored = staticmethod(
        partial(filterfalse, operator.attrgetter("ignored")))

    # this *has* to be a property, else it creates a cycle.
    parent = property(lambda s:s)

    def __init__(self):
        self.events = []

    def __str__(self):
        return 'resolver stack:\n  %s' % '\n  '.join(str(x) for x in self)

    def __repr__(self):
        return '<%s: %r>' % (self.__class__.__name__,
            tuple(repr(x) for x in self))

    def add_frame(self, mode, atom, choices, dbs, start_point, drop_cycles, vdb_limited=False):
        if not self:
            parent = self
        else:
            parent = self[-1]
        frame = self.frame_klass(parent, mode, atom, choices, dbs, start_point,
            self.depth + 1, drop_cycles, vdb_limited=vdb_limited)
        self.append(frame)
        return frame

    def add_event(self, event):
        if not self:
            self.events.append(event)
        else:
            self[-1].events.append(event)

    def pop_frame(self, result):
        frame = self.pop()
        frame.succeeded = bool(result)
        frame.parent.events.append(frame)

    def slot_cycles(self, trg_frame, **kwds):
        pkg = trg_frame.current_pkg
        slot = pkg.slot
        key = pkg.key
        kwds['skip_trg_frame'] = True
        return (frame for frame in self._cycles(trg_frame, **kwds)
            if key == frame.current_pkg.key and slot == frame.current_pkg.slot)

    def _cycles(self, trg_frame, start=0, reverse=False, skip_trg_frame=True):
        if reverse:
            i = self._filter_ignored(reversed(self))
        else:
            i = self._filter_ignored(self)
        if start != 0:
            i = islice(i, start, None)
        if skip_trg_frame:
            return (frame for frame in i if frame is not trg_frame)
        return i

    def index(self, frame, start=0, stop=None):
        i = self
        if start != 0 or stop is not None:
            i = slice(i, start, stop)

        for idx, x in enumerate(self):
            if x == frame:
                return idx + start
        return -1


class merge_plan:

    vdb_restrict = packages.PackageRestriction("repo.livefs", values.EqualityMatch(True))

    def __init__(self, dbs, per_repo_strategy, global_strategy=None,
                 depset_reorder_strategy=None, process_built_depends=False,
                 drop_cycles=False, debug=False, debug_handle=None):
        if debug:
            if debug_handle is None:
                debug_handle = sys.stdout
            self._dprint = partial(dprint, debug_handle)
        else:
            # don't run debug func when debugging is disabled
            self._dprint = lambda *args, **kwargs: None

        if not isinstance(dbs, (util.RepositoryGroup, list, tuple)):
            dbs = [dbs]

        if global_strategy is None:
            global_strategy = self.default_global_strategy

        if depset_reorder_strategy is None:
            depset_reorder_strategy = self.default_depset_reorder_strategy

        self.depset_reorder = depset_reorder_strategy
        self.all_raw_dbs = [misc.caching_repo(x, per_repo_strategy) for x in dbs]
        self.all_dbs = global_strategy(self.all_raw_dbs)
        self.default_dbs = self.all_dbs

        self.state = state.plan_state()
        vdb_state_filter_restrict = MutableContainmentRestriction(self.state.vdb_filter)
        self.livefs_dbs = multiplex.tree(
            *[filtered.tree(x, vdb_state_filter_restrict)
                for x in self.all_raw_dbs if x.livefs])

        self.insoluble = set()
        self.vdb_preloaded = False
        self._ensure_livefs_is_loaded = \
            self._ensure_livefs_is_loaded_nonpreloaded
        self.drop_cycles = drop_cycles
        self.process_built_depends = process_built_depends
        self._debugging = debug
        if debug:
            self._rec_add_atom = partial(self._stack_debugging_rec_add_atom,
                self._rec_add_atom)
            self._debugging_depth = 0
            self._debugging_drop_cycles = False

    @property
    def forced_restrictions(self):
        return frozenset(self.state.forced_restrictions)

    def reset(self, point=0):
        self.state.backtrack(point)

    def notify_starting_mode(self, mode, stack):
        if mode == "pdepend":
            mode = 'prdepends'
        self._dprint(
            "%s:%s%s: started: %s",
            (mode, ' ' * ((stack.current_frame.depth * 2) + 12 - len(mode)),
                stack.current_frame.atom,
                stack.current_frame.choices.current_pkg)
            )

    def notify_trying_choice(self, stack, atom, choices):
        self._dprint(
            "choose for %s%s, %s",
            (stack.depth *2*" ", atom, choices.current_pkg))
        stack.add_event(('inspecting', choices.current_pkg))

    def notify_choice_failed(self, stack, atom, choices, msg, msg_args=()):
        stack[-1].events.append(("choice", str(choices.current_pkg), False, msg % msg_args))
        if msg:
            msg = ': %s' % (msg % msg_args)
        self._dprint(
            "choice for %s%s, %s failed%s",
            (stack.depth * 2 * ' ', atom, choices.current_pkg, msg))

    def notify_choice_succeeded(self, stack, atom, choices, msg='', msg_args=()):
        stack[-1].events.append(("choice", str(choices.current_pkg), True, msg))
        if msg:
            msg = ': %s' % (msg % msg_args)
        self._dprint(
            "choice for %s%s, %s succeeded%s",
            (stack.depth * 2 * ' ', atom, choices.current_pkg, msg))

    def notify_viable(self, stack, atom, viable, msg='', pre_solved=False):
        t_viable = viable and "processing" or "not viable"
        if pre_solved and viable:
            t_viable = "pre-solved"
        t_msg = msg and (" "+msg) or ''
        s=''
        if stack:
            s = " for %s " % (stack[-1].atom)
        self._dprint(
            "%s%s%s%s%s",
            (t_viable.ljust(13), "  "*stack.depth, atom, s, t_msg))
        stack.add_event(("viable", viable, pre_solved, atom, msg))

    def load_vdb_state(self):
        for pkg in self.livefs_dbs:
            self._dprint("inserting %s", (pkg,), "vdb")
            ret = self.add_atom(pkg.versioned_atom)
            self._dprint("insertion of %s: %s", (pkg, ret), "vdb")
            if ret:
                raise Exception(
                    "couldn't load vdb state, %s %s" %
                    (pkg.versioned_atom, ret))
        self.vdb_preloaded = True
        self._ensure_livefs_is_loaded = \
            self._ensure_livefs_is_loaded_preloaded

    def add_atoms(self, restricts, finalize=False):
        if restricts:
            stack = resolver_stack()
            for restrict in restricts:
                state.add_hardref_op(restrict).apply(self.state)
            dbs = self.default_dbs
            for restrict in restricts:
                ret = self._add_atom(restrict, stack, dbs)
                if ret:
                    return ret
        if finalize:
            # note via this being outside the recursion, backtracking
            # is excluded... inline it somehow.
            self.process_finalize()
        return ()

    def process_finalize(self):
        pass

    def add_atom(self, atom):
        """add an atom, recalculating as necessary.

        :return: the last unresolvable atom stack if a solution can't be found,
            else returns None if the atom was successfully added.
        """
        return self.add_atoms([atom])

    def _add_atom(self, atom, stack, dbs):
        ret = self._rec_add_atom(atom, stack, dbs)
        if ret:
            self._dprint("failed- %s", ret)
            return ret, stack.events[-1]
        return ()

    def _stack_debugging_rec_add_atom(self, func, atom, stack, dbs, **kwds):
        current = len(stack)
        cycles = kwds.get('drop_cycles', False)
        reset_cycles = False
        if cycles and not self._debugging_drop_cycles:
            self._debugging_drop_cycles = reset_cycles = True
        if not reset_cycles:
            self._debugging_depth += 1

        assert current == self._debugging_depth -1
        ret = func(atom, stack, dbs, **kwds)
        assert current == len(stack)
        assert current == self._debugging_depth -1
        if not reset_cycles:
            self._debugging_depth -= 1
        else:
            self._debugging_drop_cycles = False
        return ret

    def _rec_add_atom(self, atom, stack, dbs, mode="none", drop_cycles=False):
        """Add an atom.

        :return: False on no issues (inserted succesfully),
            else a list of the stack that screwed it up.
        """
        assert hasattr(dbs, 'itermatch')
        limit_to_vdb = dbs == self.livefs_dbs

        matches = self._viable(stack, mode, atom, dbs, drop_cycles, limit_to_vdb)
        if matches is None:
            stack.pop_frame(False)
            return [atom]
        elif matches is True:
            stack.pop_frame(True)
            return None
        choices, matches = matches

        depth = stack.depth

        if stack:
            if limit_to_vdb:
                self._dprint(
                    "processing   %s%s  [%s]; mode %s vdb bound",
                    (depth*2*" ", atom, stack[-1].atom, mode))
            else:
                self._dprint(
                    "processing   %s%s  [%s]; mode %s",
                    (depth*2*" ", atom, stack[-1].atom, mode))
        else:
            self._dprint("processing   %s%s", (depth*2*" ", atom))

        ret = self.check_for_cycles(stack, stack.current_frame)
        if ret is not True:
            stack.pop_frame(ret is None)
            return ret

        failures = []

        debugging = self._debugging
        last_state = None
        while choices:
            if debugging:
                new_state = choices.state
                if last_state == new_state:
                    raise AssertionError(
                        "no state change detected, "
                        "old %r != new %r\nchoices(%r)\ncurrent(%r)\n"
                        "bdepend(%r)\ndepend(%r)\nrdepend(%r)\npdepend(%r)\n"
                        "idepend(%r)" % (
                            last_state, new_state, tuple(choices.matches), choices.current_pkg,
                            choices.bdepend, choices.depend, choices.rdepend, choices.pdepend,
                            choices.idepend,
                        )
                    )
                last_state = new_state
            additions = []

            self.notify_trying_choice(stack, atom, choices)

            if not choices.current_pkg.built or self.process_built_depends:
                new_additions, failures = self.process_dependencies_and_blocks(
                    stack, choices, 'depend', atom, depth)
                if failures:
                    continue
                additions += new_additions

                new_additions, failures = self.process_dependencies_and_blocks(
                    stack, choices, 'bdepend', atom, depth)
                if failures:
                    continue
                additions += new_additions

            new_additions, failures = self.process_dependencies_and_blocks(
                stack, choices, 'rdepend', atom, depth)
            if failures:
                continue
            additions += new_additions

            # TODO: do we need a conditional for merging a pkg here?
            new_additions, failures = self.process_dependencies_and_blocks(
                stack, choices, 'idepend', atom, depth)
            if failures:
                continue
            additions += new_additions

            l = self.insert_choice(atom, stack, choices)
            if l is False:
                # this means somehow the node already slipped in.
                # so we exit now, we are satisfied
                self.notify_choice_succeeded(
                    stack, atom, choices,
                    "already exists in the state plan")
                stack.pop_frame(True)
                return None
            elif l is not None:
                # failure.
                self.notify_choice_failed(
                    stack, atom, choices,
                    "failed inserting: %s", l)
                self.state.backtrack(stack.current_frame.start_point)
                choices.force_next_pkg()
                continue

            new_additions, failures = self.process_dependencies_and_blocks(
                stack, choices, 'pdepend', atom, depth)
            if failures:
                continue
            additions += new_additions

            self.notify_choice_succeeded(stack, atom, choices)
            stack.pop_frame(True)
            return None

        self._dprint("no solution  %s%s", (depth*2*" ", atom))
        stack.add_event(("debug", "ran out of choices",))
        self.state.backtrack(stack.current_frame.start_point)
        # saving roll.  if we're allowed to drop cycles, try it again.
        # this needs to be *far* more fine grained also. it'll try
        # regardless of if it's a cycle issue
        if not drop_cycles and self.drop_cycles:
            stack.add_event(("cycle", stack.current_frame, "trying to drop any cycles"),)
            self._dprint(
                "trying saving throw for %s ignoring cycles",
                atom, "cycle")
            # note everything is retored to a pristine state prior also.
            stack[-1].ignored = True
            l = self._rec_add_atom(atom, stack, dbs,
                mode=mode, drop_cycles=True)
            if not l:
                stack.pop_frame(True)
                return None
        stack.pop_frame(False)
        return [atom] + failures

    def _viable(self, stack, mode, atom, dbs, drop_cycles, limit_to_vdb):
        """
        internal function to discern if an atom is viable, returning
        the choicepoint/matches iterator if viable.

        :param stack: current stack
        :type stack: :obj:`resolver_stack`
        :param mode: type of dependency (depend/rdepend)
        :type mode: str
        :param atom: atom for the current package
        :type atom: :obj:`pkgcore.ebuild.atom.atom`
        :param dbs: db list to walk
        :param drop_cycles: boolean controlling whether to drop dep cycles
        :param limit_to_vdb: boolean controlling considering pkgs only from the vdb
        :return: 3 possible; None (not viable), True (presolved),
          :obj:`caching_iter` (not solved, but viable), :obj:`choice_point`
        """
        choices = ret = None
        if atom in self.insoluble:
            ret = ((False, "globally insoluble"),{})
            matches = ()
        else:
            matches = self.state.match_atom(atom)
            if matches:
                ret = ((True,), {"pre_solved":True})
            else:
                # not in the plan thus far.
                matches = caching_iter(dbs.itermatch(atom))
                if matches:
                    choices = choice_point(atom, matches)
                    # ignore what dropped out, at this juncture we don't care.
                    choices.reduce_atoms(self.insoluble)
                    if not choices:
                        # and was intractable because it has a hard dep on an
                        # unsolvable atom.
                        ret = ((False, "pruning of insoluble deps "
                            "left no choices"), {})
                else:
                    ret = ((False, "no matches"), {})

        if choices is None:
            choices = choice_point(atom, matches)

        stack.add_frame(mode, atom, choices, dbs,
            self.state.current_state, drop_cycles, vdb_limited=limit_to_vdb)

        if not limit_to_vdb and not matches:
            self.insoluble.add(atom)
        if ret is not None:
            self.notify_viable(stack, atom, *ret[0], **ret[1])
            if ret[0][0] == True:
                state.add_backref_op(choices, choices.current_pkg).apply(self.state)
                return True
            return None
        return choices, matches

    def check_for_cycles(self, stack, cur_frame):
        """Check the current stack for cyclical issues.

        :param stack: current stack, a :obj:`resolver_stack` instance
        :param cur_frame: current frame, a :obj:`resolver_frame` instance
        :return: True if no issues and resolution should continue, else the
            value to return after collapsing the calling frame
        """
        force_vdb = False
        for frame in stack.slot_cycles(cur_frame, reverse=True):
            if not any(f.mode == 'pdepend' for f in
                islice(stack, stack.index(frame), stack.index(cur_frame))):
                # exact same pkg.
                if frame.mode in ('bdepend', 'depend'):
                    # ok, we *must* go vdb if not already.
                    if frame.current_pkg.repo.livefs:
                        if cur_frame.current_pkg.repo.livefs:
                            return None
                        # force it to vdb.
                    if cur_frame.current_pkg.repo.livefs:
                        return True
                    elif cur_frame.current_pkg == frame.current_pkg and \
                        cur_frame.mode == 'pdepend':
                        # if non vdb and it's a post_rdeps cycle for the cur
                        # node, exempt it; assuming the stack succeeds,
                        # it's satisfied
                        return True
                    force_vdb = True
                    break
                else:
                    # should be doing a full walk of the cycle here, seeing
                    # if an rdep becomes a dep.
                    return None
                # portage::gentoo -> rysnc -> portage::vdb; let it process it.
                return True
            # only need to look at the most recent match; reasoning is simple,
            # logic above forces it to vdb if needed.
            break
        if not force_vdb:
            return True
        # we already know the current pkg isn't livefs; force livefs to
        # sidestep this.
        cur_frame.parent.events.append(("cycle", cur_frame, "limiting to vdb"))
        cur_frame.ignored = True
        return self._rec_add_atom(cur_frame.atom, stack,
            self.livefs_dbs, mode=cur_frame.mode,
            drop_cycles = cur_frame.drop_cycles)

    def process_dependencies_and_blocks(self, stack, choices, attr,
                                        atom=None, depth=None):
        if atom is None:
            atom = stack.current_frame.atom
        if depth is None:
            depth = stack.depth
        depset = self.depset_reorder(getattr(choices, attr), attr)
        l = self.process_dependencies(stack, choices, attr, depset, atom)
        if len(l) == 1:
            self._dprint(
                "resetting for %s%s because of %s: %s",
                (depth*2*" ", atom, attr, l[0]))
            self.state.backtrack(stack.current_frame.start_point)
            return [], l[0]

        additions = l[0]
        return additions, []

    def process_dependencies(self, stack, choices, mode, depset, atom):
        failure = []
        additions, blocks, = [], []
        cur_frame = stack.current_frame
        self.notify_starting_mode(mode, stack)
        for potentials in depset:
            failure = []
            for or_node in potentials:
                if or_node.blocks:
                    failure = self.process_blocker(stack, choices, or_node, mode, atom)
                    if not failure:
                        blocks.append(or_node)
                        break
                else:
                    failure = self._rec_add_atom(or_node, stack,
                        cur_frame.dbs, mode=mode,
                        drop_cycles=cur_frame.drop_cycles)
                    if not failure:
                        additions.append(or_node)
                        break
                    # XXX this is whacky tacky fantastically crappy
                    # XXX kill it; purpose seems... questionable.
                    if cur_frame.drop_cycles:
                        self._dprint(
                            "%s level cycle: %s: "
                            "dropping cycle for %s from %s",
                            (mode, cur_frame.atom, or_node, cur_frame.current_pkg),
                            "cycle")
                        failure = None
                        break

                if cur_frame.reduce_solutions(or_node):
                    # pkg changed.
                    return [failure]
                continue
            else: # didn't find any solutions to this or block.
                cur_frame.reduce_solutions(potentials)
                return [potentials]
        else: # all potentials were usable.
            return additions, blocks

    def process_blocker(self, stack, choices, blocker, mode, atom):
        ret = self.insert_blockers(stack, choices, [blocker])
        if ret is None:
            return []
        self.notify_choice_failed(
            stack, atom, choices,
            "%s blocker: %s conflicts w/ %s", (mode, ret[0], ret[1]))
        return [ret[0]]

    def _ensure_livefs_is_loaded_preloaded(self, restrict):
        return

    def _ensure_livefs_is_loaded_nonpreloaded(self, restrict):
        # do a trick to make the resolver now aware of vdb pkgs if needed
        # check for any matches; none, try and insert vdb nodes.
        l = self.state.match_atom(restrict)
        if not l:
            # hmm. ok... no conflicts, so we insert in vdb matches
            # to trigger a replace instead of an install
            for pkg in self.livefs_dbs.itermatch(restrict):
                self._dprint("inserting vdb node for %s %s", (restrict, pkg))
                c = choice_point(restrict, [pkg])
                state.add_op(c, c.current_pkg, force=True).apply(self.state)

    def insert_choice(self, atom, stack, choices):
        # first, check for conflicts.
        # lil bit fugly, but works for the moment
        if not choices.current_pkg.repo.livefs:
            self._ensure_livefs_is_loaded(choices.current_pkg.slotted_atom)
        conflicts = state.add_op(choices, choices.current_pkg).apply(self.state)
        if conflicts:
            # this means in this branch of resolution, someone slipped
            # something in already. cycle, basically.
            # hack.  see if what was inserted is enough for us.

            # this is tricky... if it's the same node inserted
            # (cycle), then we ignore it; this does *not* perfectly
            # behave though, doesn't discern between repos.

            # Note that virtual pkg conflicts are skipped since it's assumed
            # they are injected.
            virtual = (any(not getattr(x, 'package_is_real', True) for x in conflicts)
                       or not choices.current_pkg.package_is_real)
            if (virtual or (len(conflicts) == 1 and conflicts[0] == choices.current_pkg and
                    (conflicts[0].repo.livefs == choices.current_pkg.repo.livefs and
                    atom.match(conflicts[0])))):
                # early exit. means that a cycle came about, but exact
                # same result slipped through.
                return False

            self._dprint(
                "was trying to insert atom '%s' pkg '%s',\nbut '[%s]' exists already",
                (atom, choices.current_pkg, ", ".join(map(str, conflicts))))

            try_rematch = False
            if any(True for x in conflicts if isinstance(x, restriction.base)):
                # blocker was caught
                try_rematch = True
            elif not any(True for x in conflicts if not self.vdb_restrict.match(x)):
                # vdb entry, replace.
                if self.vdb_restrict.match(choices.current_pkg):
                    # we're replacing a vdb entry with a vdb entry?  wtf.
                    print("internal weirdness spotted- vdb restrict matches, "
                          "but current doesn't, bailing")
                    raise Exception("internal weirdness- vdb restrict matches ",
                                    "but current doesn't. bailing- run w/ --debug")
                conflicts = state.replace_op(choices, choices.current_pkg).apply(self.state)
                if not conflicts:
                    self._dprint(
                        "replacing vdb entry for '%s' with pkg '%s'",
                        (atom, choices.current_pkg))

            else:
                try_rematch = True
            if try_rematch:
                # XXX: this block looks whacked.  figure out what it's up to.
                l2 = self.state.match_atom(atom)
                if l2 == [choices.current_pkg]:
                    # stop resolution.
                    conflicts = False
                elif l2:
                    # potentially need to do some form of cleanup here.
                    conflicts = False
        else:
            conflicts = None
        return conflicts

    def generate_mangled_blocker(self, choices, blocker):
        """converts a blocker into a "cannot block ourself" block"""
        # note the second Or clause is a bit loose; allows any version to
        # slip through instead of blocking everything that isn't the
        # parent pkg
        if blocker.category != 'virtual':
            return blocker
        return packages.AndRestriction(blocker,
            packages.PackageRestriction("provider.key",
                values.StrExactMatch(choices.current_pkg.key),
                negate=True, ignore_missing=True))

    def insert_blockers(self, stack, choices, blocks):
        # level blockers.
        was_livefs = choices.current_pkg.repo.livefs
        for x in blocks:
            if not was_livefs:
                self._ensure_livefs_is_loaded(x)

            rewrote_blocker = self.generate_mangled_blocker(choices, x)
            l = self.state.add_blocker(choices, rewrote_blocker, key=x.key)
            if l:
                # blocker caught something. yay.
                self._dprint(
                    "%s blocker %s hit %s for atom %s pkg %s",
                    (stack[-1].mode, x, l, stack[-1].atom, choices.current_pkg))
                if x.weak_blocker:
                    # note that we use the top frame of the stacks' dbs; this
                    # is to allow us to upgrade as needed.
                    # For this to match, it's *only* possible if the blocker is resolved
                    # since the limiter is already in place.
                    result = self._rec_add_atom(packages.KeyedAndRestriction(
                        restriction.Negate(x), _atom.atom(x.key), key=x.key), stack, stack[0].dbs)
                    if not result:
                        # ok, inserted a new version.  did it take care of the conflict?
                        # it /may/ not have, via filling a different slot...
                        result = self.state.match_atom(x)
                        if not result:
                            # ignore the blocker, we resolved past it.
                            continue
                return x, l
        return None

    def free_caches(self):
        for repo in self.all_raw_dbs:
            repo.clear()

    # selection strategies for atom matches

    def default_depset_reorder_strategy(self, depset, mode):
        for or_block in depset:
            vdb = []
            non_vdb = []
            if len(or_block) == 1:
                yield or_block
                continue
            for atom in or_block:
                if atom.blocks:
                    non_vdb.append(atom)
                elif self.state.match_atom(atom):
                    vdb.append(atom)
                elif atom in self.livefs_dbs:
                    vdb.append(atom)
                else:
                    non_vdb.append(atom)
            if vdb:
                yield vdb + non_vdb
            else:
                yield or_block

    @staticmethod
    def default_global_strategy(dbs, atom):
        return chain(*[repo.itermatch(atom) for repo in dbs])

    @staticmethod
    def just_livefs_dbs(dbs):
        return (r for r in dbs if r.livefs)

    @staticmethod
    def just_nonlivefs_dbs(dbs):
        return (r for r in dbs if not r.livefs)

    @classmethod
    def prefer_livefs_dbs(cls, dbs, just_vdb=None):
        """
        :param dbs: db list to walk
        :param just_vdb: if None, no filtering; if True, just vdb, if False,
          non-vdb only
        :return: yields repos in requested ordering
        """
        return chain(cls.just_livefs_dbs(dbs), cls.just_nonlivefs_dbs(dbs))

    @classmethod
    def prefer_nonlivefs_dbs(cls, dbs, just_vdb=None):
        """
        :param dbs: db list to walk
        :param just_vdb: if None, no filtering; if True, just vdb, if False,
          non-vdb only
        :return: yields repos in requested ordering
        """
        return chain(cls.just_nonlivefs_dbs(dbs), cls.just_livefs_dbs(dbs))

    @classmethod
    def prefer_highest_version_strategy(cls, dbs):
        return misc.multiplex_sorting_repo(
            highest_iter_sort, cls.prefer_livefs_dbs(dbs))

    @staticmethod
    def prefer_lowest_version_strategy(dbs):
        return misc.multiplex_sorting_repo(lowest_iter_sort, dbs)

    @classmethod
    def prefer_downgrade_version_strategy(cls, restrict, dbs):
        return misc.multiplex_sorting_repo(
            partial(downgrade_iter_sort, restrict),
            cls.prefer_nonlivefs_dbs(dbs))

    @classmethod
    def prefer_reuse_strategy(cls, dbs):
        return multiplex.tree(
            misc.multiplex_sorting_repo(
                highest_iter_sort, cls.just_livefs_dbs(dbs)),
            misc.multiplex_sorting_repo(
                highest_iter_sort, cls.just_nonlivefs_dbs(dbs)),
        )



"""system cleaning utility"""

import argparse
import errno
import glob
import os
import re
import shutil
import sys
import time
from collections import defaultdict
from functools import partial
from itertools import chain

from snakeoil.mappings import DictMixin
from snakeoil.osutils import listdir_dirs, listdir_files, pjoin
from snakeoil.sequences import iflatten_instance, split_negations
from snakeoil.strings import pluralism

from ..ebuild import atom as atom_mod
from ..ebuild.domain import domain as domain_cls
from ..repository import multiplex
from ..repository.util import SimpleTree, get_virtual_repos
from ..restrictions import boolean, packages
from ..util import parserestrict
from ..util.commandline import ArgumentParser, StoreRepoObject, convert_to_restrict

argparser = ArgumentParser(description=__doc__, script=(__file__, __name__))
subparsers = argparser.add_subparsers(description='cleaning applets')
@argparser.bind_parse_priority(10)
def _initialize_opts(namespace):
    namespace.restrict = []
    namespace.file_filters = Filters()

shared_opts = ArgumentParser(suppress=True)
cleaning_opts = shared_opts.add_argument_group('generic cleaning options')
cleaning_opts.add_argument(
    nargs='*', dest='targets', metavar='TARGET',
    help="packages to target for cleaning")
cleaning_opts.add_argument(
    '-p', '--pretend', action='store_true',
    help='dry run without performing any changes')
cleaning_opts.add_argument(
    '-x', '--exclude', action='csv', dest='excludes', metavar='EXCLUDE',
    help='list of packages to exclude from removal')
cleaning_opts.add_argument(
    '-X', '--exclude-file', type=argparse.FileType('r'),
    help='path to exclusion file')
cleaning_opts.add_argument(
    '-S', '--pkgsets', action='csv_negations', metavar='PKGSET',
    help='list of pkgsets to include or exclude from removal')
@shared_opts.bind_parse_priority(20)
def _setup_shared_opts(namespace):
    namespace.exclude_restrict = None
    exclude_restrictions = []

    if namespace.pkgsets:
        disabled, enabled = namespace.pkgsets
        unknown_sets = set(disabled + enabled).difference(namespace.config.pkgset)
        if unknown_sets:
            argparser.error("unknown set%s: %s (available sets: %s)" % (
                pluralism(unknown_sets),
                ', '.join(sorted(map(repr, unknown_sets))),
                ', '.join(sorted(namespace.config.pkgset))))
        for s in set(disabled):
            exclude_restrictions.extend(namespace.config.pkgset[s])
        for s in set(enabled):
            namespace.restrict.append(boolean.OrRestriction(*namespace.config.pkgset[s]))

    # handle command line and file excludes
    excludes = namespace.excludes if namespace.excludes is not None else []
    if namespace.exclude_file is not None:
        excludes.extend(namespace.exclude_file.read().split('\n'))
    if excludes:
        exclude_restrictions.extend(convert_to_restrict(excludes, default=None))

    if exclude_restrictions:
        namespace.restrict.append(
            boolean.OrRestriction(negate=True, *exclude_restrictions))
        namespace.exclude_restrict = boolean.OrRestriction(*exclude_restrictions)


def parse_time(s):
    # simple approximations, could use dateutil for exact deltas
    units = {'s': 1}
    units['min'] = units['s'] * 60
    units['h'] = units['min'] * 60
    units['d'] = units['h'] * 24
    units['w'] = units['d'] * 7
    units['m'] = units['d'] * 30
    units['y'] = units['d'] * 365

    date = re.match(r'^(\d+)(%s)$' % '|'.join(units.keys()), s)
    if date:
        value = int(date.group(1))
        unit = date.group(2)
    else:
        raise argparse.ArgumentTypeError(
            f"invalid date: {s!r} (valid units: {' ,'.join(units.keys())})")
    return time.time() - (value * units[unit])


def parse_size(s):
    units = {
        'B': 1,
        'K': 1024,
        'M': 1024**2,
        'G': 1024**3,
    }

    size = re.match(r'^(\d+)([%s])$' % ''.join(units.keys()), s)
    if size:
        value = int(size.group(1))
        unit = size.group(2)
    else:
        raise argparse.ArgumentTypeError(
            f"invalid size: {s!r} (valid units: {' ,'.join(units.keys())})")
    return value * units[unit]


file_opts = ArgumentParser(suppress=True)
file_cleaning_opts = file_opts.add_argument_group('file cleaning options')
file_cleaning_opts.add_argument(
    '-m', '--modified', metavar='TIME', type=parse_time,
    help='skip files that have been modified since a given time',
    docs="""
        Don't remove files that have been modified since a given time. For
        example, to skip files newer than a year use "1y" as an argument to this
        option.  this option.

        Supported units are y, m, w, and d, and s representing years, months,
        weeks, days, and seconds, respectively.
    """)
file_cleaning_opts.add_argument(
    '-s', '--size', metavar='SIZE', type=parse_size,
    help='skip files bigger than a given size',
    docs="""
        Don't remove files bigger than a given size.  For example, to skip
        files larger than 100 megabytes use "100M" as an argument to this
        option.

        Supported units are B, K, M, and G representing bytes, kilobytes,
        megabytes, and gigabytes, respectively.
    """)
@file_opts.bind_parse_priority(20)
def _setup_file_opts(namespace):
    if namespace.modified is not None:
        namespace.file_filters.append(lambda x: os.stat(x).st_mtime < namespace.modified)
    if namespace.size is not None:
        namespace.file_filters.append(lambda x: os.stat(x).st_size < namespace.size)


repo_opts = ArgumentParser(suppress=True)
repo_cleaning_opts = repo_opts.add_argument_group('repo cleaning options')
repo_cleaning_opts.add_argument(
    '-I', '--installed', action='store_true', dest='exclude_installed',
    help='skip files for packages that are currently installed')
repo_cleaning_opts.add_argument(
    '-E', '--exists', action='store_true', dest='exclude_exists',
    help='skip files for packages that relate to ebuilds in the tree')
repo_cleaning_opts.add_argument(
    '-f', '--fetch-restricted', action='store_true', dest='exclude_fetch_restricted',
    help='skip fetch-restricted files')
repo_cleaning_opts.add_argument(
    "-r", "--repo", help="target repository",
    action=StoreRepoObject,
    docs="""
        Target repository to search for matches. If no repo is specified all
        relevant repos are used.
    """)


@argparser.bind_parse_priority(30)
def _setup_restrictions(namespace):
    repo = namespace.domain.all_source_repos_raw
    target_restrictions = []

    # If no targets are passed, create a restriction from the current working
    # directory if inside a known repo.
    cwd = os.getcwd()
    if not namespace.targets and cwd in repo:
        namespace.targets = [cwd]

    for target in namespace.targets:
        try:
            target_restrictions.append(parserestrict.parse_match(target))
        except parserestrict.ParseError as e:
            if os.path.exists(target):
                try:
                    restrict = repo.path_restrict(target)
                    # toss the repo restriction, keep the rest
                    target_restrictions.append(boolean.AndRestriction(*restrict[1:]))
                except ValueError as e:
                    argparser.error(e)
            else:
                argparser.error(e)

    if target_restrictions:
        namespace.restrict.append(boolean.OrRestriction(*target_restrictions))
    if namespace.restrict:
        namespace.restrict = boolean.AndRestriction(*namespace.restrict)


class _UnfilteredRepos(DictMixin):
    """Generate custom, unfiltered repos on demand."""

    _supported_attrs = {
        'pkg_masks', 'pkg_unmasks', 'pkg_accept_keywords', 'pkg_keywords',
    }

    def __init__(self, domain):
        self.domain = domain
        self.unfiltered_repos = {}

    def __getitem__(self, key):
        if key not in self._supported_attrs:
            raise KeyError

        try:
            return self.unfiltered_repos[key]
        except KeyError:
            repos = []
            kwargs = {key: ()}
            for repo in self.domain.ebuild_repos_unfiltered:
                repos.append(self.domain.filter_repo(repo, **kwargs))
            unfiltered_repo = multiplex.tree(*repos)
            self.unfiltered_repos[key] = unfiltered_repo
            return unfiltered_repo

    def keys(self):
        return self.unfiltered_repo.keys()

    def values(self):
        return self.unfiltered_repo.values()

    def items(self):
        return self.unfiltered_repo.items()


config = subparsers.add_parser(
   'config', parents=(shared_opts,),
   description='remove config file settings')
@config.bind_main_func
def config_main(options, out, err):
    domain = options.domain
    installed_repos = domain.all_installed_repos
    all_repos_raw = domain.all_repos_raw
    all_ebuild_repos = domain.all_ebuild_repos

    # proxy to create custom, unfiltered repos
    unfiltered_repos = _UnfilteredRepos(domain)

    def iter_restrict(iterable):
        for x in iterable:
            restrict = x[0]
            if (options.exclude_restrict is None or
                    not options.exclude_restrict.match(restrict)):
                yield restrict, list(x)

    domain_attrs = (
        'pkg_masks', 'pkg_unmasks', 'pkg_keywords', 'pkg_accept_keywords',
        'pkg_licenses', 'pkg_use', 'pkg_env',
    )

    attrs = {}
    for name in domain_attrs:
        # call jitted attr funcs directly to provide debug data
        func = getattr(domain_cls, name).function
        # filter excluded, matching restricts from the data stream
        attrs[name] = iter_restrict(func(domain, debug=True))

    changes = defaultdict(lambda: defaultdict(list))
    for name, iterable in attrs.items():
        for restrict, item in iterable:
            path, lineno, line = item.pop(), item.pop(), item.pop()
            if not all_repos_raw.match(restrict):
                changes['unavailable'][path].append((line, lineno, str(restrict)))
                continue

            if not installed_repos.match(restrict):
                changes['uninstalled'][path].append((line, lineno, str(restrict)))

            if name in unfiltered_repos:
                filtered_pkgs = all_ebuild_repos.match(restrict)
                unfiltered_pkgs = unfiltered_repos[name].match(restrict)
                if filtered_pkgs == unfiltered_pkgs:
                    changes[f'unnecessary_{name}'][path].append((line, lineno, str(restrict)))
            elif name == 'pkg_use':
                atom, use = item

                # find duplicates
                use_sets = [set(), set()]
                disabled, enabled = use_sets
                duplicates = set()
                for i, data in enumerate(split_negations(use)):
                    for u in data:
                        if u in use_sets[i]:
                            duplicates.add(u)
                        use_sets[i].add(u)
                if duplicates:
                    changes['duplicate_use'][path].append(
                        (line, lineno, ', '.join(duplicates)))

                # find conflicts
                conflicting = enabled & disabled
                if conflicting:
                    changes['conflicting_use'][path].append(
                        (line, lineno, ', '.join(conflicting)))

                # find unknowns
                pkgs = all_repos_raw.match(atom)
                available = {u for pkg in pkgs for u in pkg.iuse_stripped}
                unknown = (disabled - available) | (enabled - available)
                if unknown:
                    changes['unknown_use'][path].append(
                        (line, lineno, ', '.join(unknown)))

    type_mapping = {
        'unavailable': 'Unavailable package(s)',
        'uninstalled': 'Uninstalled package(s)',
        'unnecessary_pkg_masks': 'Unnecessary mask(s)',
        'unnecessary_pkg_unmasks': 'Unnecessary unmask(s)',
        'unnecessary_pkg_accept_keywords': 'Unnecessary accept keywords(s)',
        'unnecessary_pkg_keywords': 'Unnecessary keywords(s)',
        'duplicate_use': 'Duplicate use flag(s)',
        'conflicting_use': 'Conflicting use flag(s)',
        'unknown_use': 'Nonexistent use flag(s)',
    }

    for t, paths in changes.items():
        out.write(f"{type_mapping[t]}:")
        for path, data in paths.items():
            out.write(f"{path}:", prefix="  ")
            for line, lineno, values in data:
                out.write(f"{values} -- line {lineno}: {line!r}", prefix="    ")
            out.write()


dist = subparsers.add_parser(
    'dist', parents=(shared_opts, file_opts, repo_opts),
    description='remove distfiles')
dist_opts = dist.add_argument_group('distfile options')
@dist.bind_final_check
def _dist_validate_args(parser, namespace):
    distdir = namespace.domain.distdir
    repo = namespace.repo
    if repo is None:
        repo = multiplex.tree(*get_virtual_repos(namespace.domain.source_repos, False))

    all_dist_files = set(os.path.basename(f) for f in listdir_files(distdir))
    target_files = set()
    installed_dist = set()
    exists_dist = set()
    excludes_dist = set()
    restricted_dist = set()

    # exclude distfiles used by installed packages -- note that this uses the
    # distfiles attr with USE settings bound to it
    if namespace.exclude_installed:
        for pkg in namespace.domain.all_installed_repos:
            installed_dist.update(iflatten_instance(pkg.distfiles))

    # exclude distfiles for existing ebuilds or fetch restrictions
    if namespace.exclude_fetch_restricted or (namespace.exclude_exists and not namespace.restrict):
        for pkg in repo:
            exists_dist.update(iflatten_instance(getattr(pkg, '_raw_pkg', pkg).distfiles))
            if 'fetch' in pkg.restrict:
                restricted_dist.update(iflatten_instance(getattr(pkg, '_raw_pkg', pkg).distfiles))

    # exclude distfiles from specified restrictions
    if namespace.exclude_restrict:
        for pkg in repo.itermatch(namespace.exclude_restrict, sorter=sorted):
            excludes_dist.update(iflatten_instance(getattr(pkg, '_raw_pkg', pkg).distfiles))

    # determine dist files for custom restrict targets
    if namespace.restrict:
        target_dist = defaultdict(lambda: defaultdict(set))
        for pkg in repo.itermatch(namespace.restrict, sorter=sorted):
            s = set(iflatten_instance(getattr(pkg, '_raw_pkg', pkg).distfiles))
            target_dist[pkg.unversioned_atom][pkg].update(s)
            if namespace.exclude_exists:
                exists_dist.update(s)

        extra_regex_prefixes = defaultdict(set)
        pkg_regex_prefixes = set()
        for catpn, pkgs in target_dist.items():
            pn_regex = r'\W'.join(re.split(r'\W', catpn.package))
            pkg_regex = re.compile(r'(%s)(\W\w+)+([\W?(0-9)+])*(\W\w+)*(\.\w+)*' % pn_regex,
                                   re.IGNORECASE)
            pkg_regex_prefixes.add(pn_regex)
            for pkg, files in pkgs.items():
                files = sorted(files)
                for f in files:
                    if (pkg_regex.match(f) or (
                            extra_regex_prefixes and
                            re.match(r'(%s)([\W?(0-9)+])+(\W\w+)*(\.\w+)+' % '|'.join(extra_regex_prefixes[catpn]), f))):
                        continue
                    else:
                        pieces = re.split(r'([\W?(0-9)+])+(\W\w+)*(\.\w+)+', f)
                        if pieces[-1] == '':
                            pieces.pop()
                        if len(pieces) > 1:
                            extra_regex_prefixes[catpn].add(pieces[0])

        if target_dist:
            regexes = []
            # build regexes to match distfiles for older ebuilds no longer in the tree
            if pkg_regex_prefixes:
                pkg_regex_prefixes_str = '|'.join(sorted(pkg_regex_prefixes))
                regexes.append(re.compile(r'(%s)(\W\w+)+([\W?(0-9)+])*(\W\w+)*(\.\w+)*' % (
                    pkg_regex_prefixes_str,)))
            if extra_regex_prefixes:
                extra_regex_prefixes_str = '|'.join(sorted(chain.from_iterable(
                    v for k, v in extra_regex_prefixes.items())))
                regexes.append(re.compile(r'(%s)([\W?(0-9)+])+(\W\w+)*(\.\w+)+' % (
                    extra_regex_prefixes_str,)))

            if regexes:
                for f in all_dist_files:
                    if any(r.match(f) for r in regexes):
                        target_files.add(f)
    else:
        target_files = all_dist_files

    # exclude files tagged for saving
    saving_files = installed_dist | exists_dist | excludes_dist | restricted_dist
    target_files.difference_update(saving_files)

    targets = (pjoin(distdir, f) for f in sorted(all_dist_files.intersection(target_files)))
    removal_func = partial(os.remove)
    namespace.remove = (
        (removal_func, f) for f in
        filter(namespace.file_filters.run, targets))


def pkg_changed(pkg, domain, attrs):
    """Determine if given attributes from a binpkg changed compared to the related ebuild."""
    try:
        repo = domain.ebuild_repos_raw[pkg.source_repository]
        ebuild_pkg = repo.match(pkg.versioned_atom)[0]
    except (KeyError, IndexError):
        # source repo doesn't exist on the system or related ebuild doesn't exist
        return False
    for attr in attrs:
        try:
            ebuild_attr = getattr(ebuild_pkg, attr)
            binpkg_attr = getattr(pkg, attr)
        except AttributeError:
            raise argparser.error(f'nonexistent attribute: {attr!r}')
        if attr.upper() in pkg.eapi.dep_keys:
            ebuild_attr = ebuild_attr.evaluate_depset(pkg.use)
        if ebuild_attr != binpkg_attr:
            return True
    return False


pkg_opts = ArgumentParser(suppress=True)
pkg_cleaning_opts = pkg_opts.add_argument_group('binpkg cleaning options')
pkg_cleaning_opts.add_argument(
    '--source-repo', metavar='REPO',
    help='remove binpkgs with matching source repo')
pkg_cleaning_opts.add_argument(
    '-b', '--bindist', action='store_true',
    help='only remove binpkgs that restrict distribution')
pkg_cleaning_opts.add_argument(
    '-c', '--changed', action='csv',
    help='comma separated list of package attributes to check for ebuild changes')
pkg = subparsers.add_parser(
    'pkg', parents=(shared_opts, file_opts, repo_opts, pkg_opts),
    description='remove binpkgs')
@pkg.bind_final_check
def _pkg_validate_args(parser, namespace):
    repo = namespace.repo
    if repo is None:
        repo = namespace.domain.all_binary_repos

    if not namespace.restrict:
        # not in a configured repo dir, remove all binpkgs
        namespace.restrict = packages.AlwaysTrue

    pkgs = (pkg for pkg in repo.itermatch(namespace.restrict))
    pkg_filters = Filters()
    if namespace.bindist:
        pkg_filters.append(lambda pkg: 'bindist' in pkg.restrict)
    if namespace.changed:
        pkg_filters.append(lambda pkg: pkg_changed(pkg, namespace.domain, namespace.changed))
    if namespace.exclude_installed:
        pkg_filters.append(lambda pkg: pkg.versioned_atom not in namespace.domain.all_installed_repos)
    if namespace.exclude_exists:
        pkg_filters.append(lambda pkg: pkg.versioned_atom not in namespace.domain.all_ebuild_repos_raw)
    if namespace.exclude_fetch_restricted:
        pkg_filters.append(lambda pkg: 'fetch' not in pkg.restrict)
    if namespace.source_repo is not None:
        pkg_filters.append(lambda pkg: namespace.source_repo == pkg.source_repository)
    pkgs = list(filter(pkg_filters.run, pkgs))
    removal_func = partial(os.remove)
    namespace.remove = (
        (removal_func, binpkg) for binpkg in
        sorted(filter(namespace.file_filters.run, (pkg.path for pkg in pkgs))))

tmp = subparsers.add_parser(
    'tmp', parents=(shared_opts,),
    description='remove tmpdir entries')
tmp_opts = tmp.add_argument_group('tmpfile options')
tmp_opts.add_argument(
    '-a', '--all', dest='wipe_all', action='store_true',
    help='wipe the entire tmpdir',
    docs="""
        Force the entire tmpdir to be wiped. Note that this overrides any
        restrictions that have been specified.
    """)
@tmp.bind_final_check
def _tmp_validate_args(parser, namespace):
    tmpdir = namespace.domain.pm_tmpdir
    dirs = ()
    files = ()

    if namespace.restrict and not namespace.wipe_all:
        # create a fake repo from tmpdir entries and pull matches from it
        pkg_map = {}
        for pkg_build_dir in glob.glob(pjoin(tmpdir, '*', '*')):
            try:
                pkg = atom_mod.atom('=' + pkg_build_dir[len(tmpdir):].lstrip(os.path.sep))
            except atom_mod.MalformedAtom:
                continue
            pkg_map.setdefault(pkg.category, {}).setdefault(pkg.package, []).append(pkg.fullver)
        repo = SimpleTree(pkg_map)

        def _remove_dir_and_empty_parent(d):
            """Remove a given directory tree and its parent directory, if empty."""
            shutil.rmtree(d)
            try:
                os.rmdir(os.path.dirname(d))
            except OSError as e:
                # POSIX specifies either ENOTEMPTY or EEXIST for non-empty dir
                # in particular, Solaris uses EEXIST in that case.
                if e.errno not in (errno.ENOTEMPTY, errno.EEXIST):
                    raise

        removal_func = partial(_remove_dir_and_empty_parent)
        dirs = ((removal_func, pjoin(tmpdir, pkg.cpvstr))
                for pkg in repo.itermatch(namespace.restrict))
    else:
        # not in a configured repo dir, remove all tmpdir entries
        dir_removal_func = partial(shutil.rmtree)
        dirs = ((dir_removal_func, pjoin(tmpdir, d)) for d in listdir_dirs(tmpdir))
        file_removal_func = partial(os.remove)
        files = ((file_removal_func, pjoin(tmpdir, f)) for f in listdir_files(tmpdir))

    namespace.remove = chain(dirs, files)


@dist.bind_main_func
@pkg.bind_main_func
@tmp.bind_main_func
def _remove(options, out, err):
    """Generic removal runner."""
    ret = 0
    if sys.stdout.isatty():
        # TODO: parallelize this
        for func, target in options.remove:
            if options.pretend and options.verbosity >= 0:
                out.write(f"Would remove {target}")
            elif options.verbosity > 0:
                out.write(f"Removing {target}")
            try:
                if not options.pretend:
                    func(target)
            except OSError as e:
                if options.verbosity >= 0:
                    err.write(f"{options.prog}: failed to remove {target!r}: {e.strerror}")
                ret = 1
                continue
    else:
        out.write('\n'.join(target for _, target in options.remove))
    return ret


class Filters:
    """Generic filtering support."""

    def __init__(self):
        self._filters = []

    def append(self, f):
        self._filters.append(f)

    @property
    def run(self):
        """Run a given object through all registered filters."""
        return lambda x: all(f(x) for f in self._filters)



import os
import pwd
from unittest import mock

import pytest
from pkgcore import os_data
from pkgcore.sync import base, git, tar
from snakeoil.process import CommandNotFound

existing_user = pwd.getpwall()[0].pw_name
existing_uid = pwd.getpwnam(existing_user).pw_uid


class TestSyncer:

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = str(tmp_path / 'repo')

    def test_split_users(self):
        o = base.Syncer(self.repo_path, "http://dar")
        assert o.uid == os.getuid()
        assert o.uri == "http://dar"

        o = base.Syncer(self.repo_path, f"http://{existing_user}::@site")
        assert o.uid == existing_uid
        assert o.uri == "http://site"

        o = base.Syncer(self.repo_path, f"http://{existing_user}::foon@site")
        assert o.uid == existing_uid
        assert o.uri == "http://foon@site"

        o = base.Syncer(self.repo_path, f"{existing_user}::foon@site")
        assert o.uid == existing_uid
        assert o.uri == "foon@site"

        with pytest.raises(base.MissingLocalUser):
            base.Syncer(self.repo_path, f"foo_nonexistent_user::foon@site")

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_usersync_disabled(self, spawn):
        o = base.Syncer(self.repo_path, f"http://foo/bar.git", usersync=False)
        o.uid == os_data.uid
        o.gid == os_data.gid

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_usersync_portage_perms(self, spawn):
        # sync uses portage perms if repo dir doesn't exist
        o = base.Syncer(self.repo_path, f"http://foo/bar.git", usersync=True)
        o.uid == os_data.portage_uid
        o.gid == os_data.portage_gid

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_usersync_repo_dir_perms(self, spawn):
        # and repo dir perms if it does exist
        with mock.patch('os.stat') as stat:
            stat.return_value = mock.Mock(st_uid=1234, st_gid=5678)
            o = base.Syncer(self.repo_path, f"http://foo/bar.git", usersync=True)
            stat.assert_called()
            assert o.uid == 1234
            assert o.gid == 5678


@mock.patch('snakeoil.process.find_binary')
class TestExternalSyncer:

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path):
        self.repo_path = str(tmp_path / 'repo')

    def test_missing_binary(self, find_binary):
        find_binary.side_effect = CommandNotFound('foo')
        with pytest.raises(base.MissingBinary):
            base.ExternalSyncer(self.repo_path, 'http://dar')

    def test_existing_binary(self, find_binary):
        # fake external syncer
        class FooSyncer(base.ExternalSyncer):
            binary = 'foo'

        # fake that the external binary exists
        find_binary.side_effect = lambda x: x

        o = FooSyncer(self.repo_path, 'http://dar')
        assert o.uri == 'http://dar'
        assert o.binary == 'foo'

    @mock.patch('snakeoil.process.spawn.spawn')
    def test_usersync(self, spawn, find_binary):
        # fake external syncer
        class FooSyncer(base.ExternalSyncer):
            binary = 'foo'

        # fake that the external binary exists
        find_binary.side_effect = lambda x: x

        o = FooSyncer(self.repo_path, 'http://dar')
        o.uid = 1234
        o.gid = 2345
        o._spawn('cmd', pipes={})
        assert spawn.call_args[1]['uid'] == o.uid
        assert spawn.call_args[1]['gid'] == o.gid


@mock.patch('snakeoil.process.find_binary', return_value='git')
@mock.patch('snakeoil.process.spawn.spawn')
class TestVcsSyncer:

    def test_basedir_perms_error(self, spawn, find_binary, tmp_path):
        syncer = git.git_syncer(str(tmp_path), 'git://blah.git')
        with pytest.raises(base.PathError):
            with mock.patch('os.stat') as stat:
                stat.side_effect = EnvironmentError('fake exception')
                syncer.sync()

    def test_basedir_is_file_error(self, spawn, find_binary, tmp_path):
        repo = tmp_path / "repo"
        repo.touch()
        syncer = git.git_syncer(str(repo), 'git://blah.git')

        # basedir gets '/' appended by default and stat errors out
        with pytest.raises(base.PathError) as excinfo:
            syncer.sync()

        # remove trailing slash from basedir and file check catches it instead
        syncer.basedir = str(repo)
        with pytest.raises(base.PathError) as excinfo:
            syncer.sync()
        assert "isn't a directory" in str(excinfo.value)

    def test_verbose_sync(self, spawn, find_binary, tmp_path):
        syncer = git.git_syncer(str(tmp_path), 'git://blah.git')
        syncer.sync(verbosity=1)
        assert '-v' == spawn.call_args[0][0][-1]
        syncer.sync(verbosity=2)
        assert '-vv' == spawn.call_args[0][0][-1]

    def test_quiet_sync(self, spawn, find_binary, tmp_path):
        syncer = git.git_syncer(str(tmp_path), 'git://blah.git')
        syncer.sync(verbosity=-1)
        assert '-q' == spawn.call_args[0][0][-1]


class TestGenericSyncer:

    def test_init(self):
        with pytest.raises(base.UriError):
            base.GenericSyncer('/', 'seriouslynotaprotocol://blah/')

        syncer = base.GenericSyncer('/', f'tar+https://blah.tar.gz')
        assert tar.tar_syncer is syncer.__class__


class TestDisabledSyncer:

    def test_init(self):
        syncer = base.DisabledSyncer('/foo/bar', f'https://blah.git')
        assert syncer.disabled
        # syncing should also be disabled
        assert not syncer.uri
        assert not syncer.sync()


class TestAutodetectSyncer:

    def test_no_syncer_detected(self, tmp_path):
        syncer = base.AutodetectSyncer(str(tmp_path))
        assert isinstance(syncer, base.DisabledSyncer)

    @mock.patch('snakeoil.process.find_binary', return_value='git')
    def test_syncer_detected(self, find_binary, tmp_path):
        d = tmp_path / '.git'
        d.mkdir()
        syncer = base.AutodetectSyncer(str(tmp_path))
        assert isinstance(syncer, git.git_syncer)



from pkgcore import plugins
from pkgcore.scripts import pplugincache
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


class CommandlineTest(TestCase, ArgParseMixin):

    _argparser = pplugincache.argparser

    has_config = False

    def test_parser(self):
        self.assertEqual([plugins], self.parse().packages)



#!/usr/bin/env python3
# Copyright 2007 Charlie Shepherd

import os
import sys
from operator import attrgetter

try:
    from pkgcore.restrictions.boolean import OrRestriction
    from pkgcore.util import commandline
except ImportError:
    print('Cannot import pkgcore!', file=sys.stderr)
    print('Verify it is properly installed and/or PYTHONPATH is set correctly.', file=sys.stderr)
    if '--debug' not in sys.argv:
        print('Add --debug to the commandline for a traceback.', file=sys.stderr)
    else:
        raise
    sys.exit(1)

argparser = commandline.ArgumentParser(color=False, version=False)
argparser.add_argument(
    'target', nargs='+', help='target package atoms')
argparser.add_argument(
    '--repo', action=commandline.StoreRepoObject,
    help='repo to use (default from domain if omitted).')
argparser.add_argument(
    '--print_type', '-t', default="cpvstr",
    choices=("slotted_atom", "versioned_atom", "cpvstr"),
    help='''type of atom to output:
                'versioned_atom' : a valid versioned atom,
                'slotted_atom'   : a valid slotted atom,
                'cpvstr'         : the cpv of the package''')


@argparser.bind_final_check
def check_args(parser, namespace):
    domain = namespace.domain
    namespace.vdb = domain.all_installed_repos
    if not namespace.repo:
        # fallback to default repo
        namespace.repo = namespace.config.get_default('repo')

    namespace.restrict = OrRestriction(
        *commandline.convert_to_restrict(namespace.target))
    namespace.outputter = attrgetter(namespace.print_type)


@argparser.bind_main_func
def main(options, out, err):
    repo = options.repo
    for built in options.vdb.itermatch(options.restrict):
        current = repo.match(built.unversioned_atom)
        if current:
            current = current[0]
            oldflags = built.iuse & built.use
            newflags = current.iuse & built.use
            if (newflags != oldflags) or (current.iuse ^ built.iuse):
                changed_flags = (oldflags ^ newflags) | (current.iuse ^ built.iuse)
                if options.verbosity > 0:
                    out.write(
                        "package %s, %d flags have changed:\n\t%s" %
                        (current.unversioned_atom, len(changed_flags), ' '.join(changed_flags)))
                else:
                    out.write(options.outputter(current))
            else:
                if options.verbosity > 0:
                    out.write("%s is the same as it was before" % current.cpvstr)


if __name__ == '__main__':
    tool = commandline.Tool(argparser)
    sys.exit(tool())



"""Plugin system, heavily inspired by twisted's plugin system."""

__all__ = ("initialize_cache", "get_plugins", "get_plugin")

# Implementation note: we have to be pretty careful about error
# handling in here since some core functionality in pkgcore uses this
# code. Since we can function without a cache we will generally be
# noisy but keep working if something is wrong with the cache.
#
# Currently we explode if something is wrong with a plugin package
# dir, but not if something prevents importing a module in it.
# Rationale is the former should be a PYTHONPATH issue while the
# latter an installed plugin issue. May have to change this if it
# causes problems.

import operator
import os.path
import sys
from collections import defaultdict, namedtuple
from importlib import import_module

from snakeoil import mappings, modules
from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.fileutils import AtomicWriteFile, readlines_ascii
from snakeoil.osutils import ensure_dirs, listdir_files, pjoin, unlink_if_exists

from . import const, os_data
from .log import logger

_plugin_data = namedtuple(
    "_plugin_data", ["key", "priority", "source", "target"])

PLUGIN_ATTR = 'pkgcore_plugins'

CACHE_HEADER = 'pkgcore plugin cache v3'
CACHE_FILENAME = 'plugincache'


def _clean_old_caches(path):
    for name in ('plugincache2',):
        try:
            unlink_if_exists(pjoin(path, name))
        except EnvironmentError as e:
            logger.error(
                "attempting to clean old plugin cache %r failed with %s",
                pjoin(path, name), e)


def sort_plugs(plugs):
    return sorted(plugs, reverse=True, key=lambda x: (x.key, x.priority, x.source))


def _process_plugins(package, sequence, filter_disabled=False):
    for plug in sequence:
        plug = _process_plugin(package, plug, filter_disabled)
        if plug is not None:
            yield plug


def _process_plugin(package, plug, filter_disabled=False):
    if isinstance(plug.target, str):
        plug = modules.load_any(plug.target)
    elif isinstance(plug.target, int):
        module = modules.load_any(plug.source)
        plugs = getattr(module, PLUGIN_ATTR, {})
        plugs = plugs.get(plug.key, [])
        if len(plugs) <= plug.target:
            logger.exception(
                "plugin cache for %s, %s, %s is somehow wrong; no item at position %s",
                package.__name__, plug.source, plug.key, plug.target)
            return None
        plug = plugs[plug.target]
    else:
        logger.error(
            "package %s, plug %s; non int, non string.  wtf?",
            package.__name__, plug)
        return None

    if filter_disabled:
        if getattr(plug, 'disabled', False):
            logger.debug("plugin %s is disabled, skipping", plug)
            return None
        f = getattr(plug, '_plugin_disabled_check', None)
        if f is not None and f():
            logger.debug("plugin %s is disabled, skipping", plug)
            return None

    return plug

def _read_cache_file(package, cache_path):
    """Read an existing cache file."""
    stored_cache = {}
    cache_data = list(readlines_ascii(cache_path, True, True, False))
    if len(cache_data) >= 1:
        if cache_data[0] != CACHE_HEADER:
            logger.warning(
                "plugin cache has a wrong header: %r, regenerating", cache_data[0])
            cache_data = []
        else:
            cache_data = cache_data[1:]
    if not cache_data:
        return {}
    try:
        for line in cache_data:
            module, mtime, entries = line.split(':', 2)
            mtime = int(mtime)
            # Needed because ''.split(':') == [''], not []
            if not entries:
                entries = set()
            else:
                entries = entries.replace(':', ',').split(',')

                if not len(entries) % 3 == 0:
                    logger.error(
                        "failed reading cache %s; entries field isn't "
                        "divisable by 3: %r", cache_path, entries)
                    continue
                entries = iter(entries)
                def f(val):
                    if val.isdigit():
                        val = int(val)
                    return val
                entries = set(
                    _plugin_data(
                        key, int(priority),
                        f'{package.__name__}.{module}', f(target))
                    for (key, priority, target) in zip(entries, entries, entries))
            stored_cache[(module, mtime)] = entries
    except IGNORED_EXCEPTIONS:
        raise
    except Exception as e:
        logger.warning("failed reading cache; exception %s, regenerating.", e)
        stored_cache.clear()

    return stored_cache

def _write_cache_file(path, data, uid=-1, gid=-1):
    """Write a new cache file."""
    cachefile = None
    try:
        try:
            cachefile = AtomicWriteFile(
                path, binary=False, perms=0o664, uid=uid, gid=gid)
            cachefile.write(CACHE_HEADER + "\n")
            for (module, mtime), plugs in sorted(data.items(), key=operator.itemgetter(0)):
                plugs = sort_plugs(plugs)
                plugs = ':'.join(f'{plug.key},{plug.priority},{plug.target}' for plug in plugs)
                cachefile.write(f'{module}:{mtime}:{plugs}\n')
            cachefile.close()
        except EnvironmentError as e:
            # We cannot write a new cache. We should log this
            # since it will have a performance impact.

            # Use error, not exception for this one: the traceback
            # is not necessary and too alarming.
            logger.error('Cannot write cache for %s: %s. '
                         'Try running pplugincache.',
                         path, e)
    finally:
        if cachefile is not None:
            cachefile.discard()


def initialize_cache(package, force=False, cache_dir=None):
    """Determine available plugins in a package.

    Writes cache files if they are stale and writing is possible.
    """
    modpath = os.path.dirname(package.__file__)
    pkgpath = os.path.dirname(os.path.dirname(modpath))
    uid = gid = -1
    mode = 0o755

    if cache_dir is None:
        if not force:
            # use user-generated caches if they exist, fallback to module cache
            if os.path.exists(pjoin(const.USER_CACHE_PATH, CACHE_FILENAME)):
                cache_dir = const.USER_CACHE_PATH
            elif os.path.exists(pjoin(const.SYSTEM_CACHE_PATH, CACHE_FILENAME)):
                cache_dir = const.SYSTEM_CACHE_PATH
                uid = os_data.portage_uid
                gid = os_data.portage_gid
                mode = 0o775
            else:
                cache_dir = modpath
        else:
            # generate module cache when running from git repo, otherwise create system/user cache
            if pkgpath == sys.path[0]:
                cache_dir = modpath
            elif os_data.uid in (os_data.root_uid, os_data.portage_uid):
                cache_dir = const.SYSTEM_CACHE_PATH
                uid = os_data.portage_uid
                gid = os_data.portage_gid
                mode = 0o775
            else:
                cache_dir = const.USER_CACHE_PATH

    # put pkgcore consumer plugins (e.g. pkgcheck) inside pkgcore cache dir
    if cache_dir in (const.SYSTEM_CACHE_PATH, const.USER_CACHE_PATH):
        chunks = package.__name__.split('.', 1)
        if chunks[0] != os.path.basename(cache_dir):
            cache_dir = pjoin(cache_dir, chunks[0])

    # package plugin cache, see above.
    package_cache = defaultdict(set)
    stored_cache_name = pjoin(cache_dir, CACHE_FILENAME)
    stored_cache = _read_cache_file(package, stored_cache_name)

    if force:
        _clean_old_caches(cache_dir)

    # Directory cache, mapping modulename to
    # (mtime, set([keys]))
    modlist = listdir_files(modpath)
    modlist = set(x for x in modlist if os.path.splitext(x)[1] == '.py'
                  and x != '__init__.py')

    cache_stale = False
    # Hunt for modules.
    actual_cache = defaultdict(set)
    mtime_cache = mappings.defaultdictkey(lambda x: int(os.path.getmtime(x)))
    for modfullname in sorted(modlist):
        modname = os.path.splitext(modfullname)[0]
        # It is an actual module. Check if its cache entry is valid.
        mtime = mtime_cache[pjoin(modpath, modfullname)]
        vals = stored_cache.get((modname, mtime))
        if vals is None or force:
            # Cache entry is stale.
            logger.debug(
                'stale because of %s: actual %s != stored %s',
                modname, mtime, stored_cache.get(modname, (0, ()))[0])
            cache_stale = True
            entries = []
            qualname = '.'.join((package.__name__, modname))
            module = import_module(qualname)
            registry = getattr(module, PLUGIN_ATTR, {})
            vals = set()
            for key, plugs in registry.items():
                for idx, plug_name in enumerate(plugs):
                    if isinstance(plug_name, str):
                        plug = _process_plugin(package, _plugin_data(key, 0, qualname, plug_name))
                    else:
                        plug = plug_name
                    if plug is None:
                        # import failure, ignore it, error already logged
                        continue
                    priority = getattr(plug, 'priority', 0)
                    if not isinstance(priority, int):
                        logger.error(
                            "ignoring plugin %s: has a non integer priority: %s",
                            plug, priority)
                        continue
                    if plug_name is plug:
                        # this means it's an object, rather than a string; store
                        # the offset.
                        plug_name = idx
                    data = _plugin_data(key, priority, qualname, plug_name)
                    vals.add(data)
        actual_cache[(modname, mtime)] = vals
        for data in vals:
            package_cache[data.key].add(data)
    if force or set(stored_cache) != set(actual_cache):
        logger.debug('updating cache %r for new plugins', stored_cache_name)
        ensure_dirs(cache_dir, uid=uid, gid=gid, mode=mode)
        _write_cache_file(stored_cache_name, actual_cache, uid=uid, gid=gid)

    return mappings.ImmutableDict((k, sort_plugs(v)) for k, v in package_cache.items())


def get_plugins(key, package=None):
    """Return all enabled plugins matching "key".

    Plugins with a C{disabled} attribute evaluating to C{True} are skipped.
    """
    if package is None:
        package = import_module('.plugins', __name__.split('.')[0])

    cache = _global_cache[package]
    for plug in _process_plugins(package, cache.get(key, ()), filter_disabled=True):
        yield plug


def get_plugin(key, package=None):
    """Get a single plugin matching this key.

    This assumes all plugins for this key have a priority attribute.
    If any of them do not the AttributeError is not stopped.

    :return: highest-priority plugin or None if no plugin available.
    """
    if package is None:
        package = import_module('.plugins', __name__.split('.')[0])

    cache = _global_cache[package]
    for plug in _process_plugins(package, cache.get(key, ()), filter_disabled=True):
        # first returned will be the highest.
        return plug
    return None


def extend_path(path, name):
    """Simpler version of the stdlib's :obj:`pkgutil.extend_path`.

    It does not support ".pkg" files, and it does not require an
    __init__.py (this is important: we want only one thing (pkgcore
    itself) to install the __init__.py to avoid name clashes).

    It also modifies the "path" list in place (and returns C{None})
    instead of copying it and returning the modified copy.
    """
    if not isinstance(path, list):
        # This could happen e.g. when this is called from inside a
        # frozen package.  Return the path unchanged in that case.
        return
    # Reconstitute as relative path.
    pname = pjoin(*name.split('.'))

    for entry in sys.path:
        if not isinstance(entry, str) or not os.path.isdir(entry):
            continue
        subdir = pjoin(entry, pname)
        # XXX This may still add duplicate entries to path on
        # case-insensitive filesystems
        if subdir not in path:
            path.append(subdir)


# Global plugin cache. Mapping of package to package cache, which is a
# mapping of plugin key to a list of module names.
_global_cache = mappings.defaultdictkey(initialize_cache)



"""
Avoid using- os data- root uid/gid, pkgcore uid/gid, etc.

This will be killed off and bound into configuration subsystem at some point
"""

__all__ = (
    "ostype", "portage_gid", "portage_uid", "root_gid", "root_uid",
    "userland", "wheelgid", "xargs",
)

import grp
import os
import pwd

ostype = os.uname()[0]

if ostype in ("Linux", "CYGWIN_NT-5.1"):
    userland = "GNU"
    xargs = os.environ["XARGS"] = "xargs -r"
    lchown = os.lchown
elif ostype == "Darwin":
    userland = "Darwin"
    xargs = os.environ["XARGS"] = "xargs"

    def lchown(*pos_args, **key_args):
        pass

elif ostype in ("FreeBSD", "OpenBSD", "NetBSD", "SunOS"):
    userland = "BSD"
    xargs = os.environ["XARGS"] = "xargs"
    lchown = os.lchown
else:
    raise Exception("Operating system unsupported, '%s'" % ostype)

uid = os.getuid()
gid = os.getgid()
# hard coding sucks.
root_uid = 0
root_gid = wheelgid = 0

try:
    wheelgid = grp.getgrnam("wheel").gr_gid
except KeyError:
    pass

# Discover the uid and gid of the portage user/group
try:
    portage_uid = pwd.getpwnam("portage").pw_uid
    portage_gid = grp.getgrnam("portage").gr_gid
    portage_user_groups = tuple(x.gr_name for x in grp.getgrall()
                                if 'portage' in x.gr_mem)
except KeyError:
    portage_uid = 0
    portage_gid = wheelgid
    portage_user_groups = []



"""
package class for buildable ebuilds
"""

__all__ = ("base", "package", "package_factory")

import os
from functools import partial
from itertools import chain
from sys import intern

from snakeoil import chksum, data_source, fileutils, klass
from snakeoil.demandload import demand_compile_regexp
from snakeoil.mappings import OrderedFrozenSet

from .. import fetch
from ..cache import errors as cache_errors
from ..log import logger
from ..package import errors as metadata_errors
from ..package import metadata
from ..package.base import DynamicGetattrSetter
from ..restrictions import boolean, values
from . import conditionals
from . import errors as ebuild_errors
from . import processor
from .atom import atom
from .eapi import get_eapi
from .misc import sort_keywords

demand_compile_regexp(
    '_EAPI_regex', r"^EAPI=(['\"]?)(?P<EAPI>[A-Za-z0-9+_.-]*)\1[\t ]*(?:#.*)?")
demand_compile_regexp(
    '_EAPI_str_regex', r"^EAPI=(['\"]?)(?P<EAPI>.*)\1")


class base(metadata.package):
    """ebuild package

    :cvar _config_wrappables: mapping of attribute to callable for
        re-evaluating attributes dependent on configuration
    """

    _config_wrappables = {
        x: klass.alias_method("evaluate_depset")
        for x in (
            "bdepend", "depend", "rdepend", "pdepend", "idepend",
            "fetchables", "license", "restrict", "required_use",
        )
    }

    __slots__ = ('_pkg_metadata_shared',)

    def _generate_depset(self, kls, key):
        return conditionals.DepSet.parse(
            self.data.pop(key, ""), kls,
            attr=key, element_func=self.eapi.atom_kls,
            transitive_use_atoms=self.eapi.options.transitive_use_atoms)

    @DynamicGetattrSetter.register
    def bdepend(self):
        if "BDEPEND" in self.eapi.metadata_keys:
            return self._generate_depset(atom, "BDEPEND")
        return conditionals.DepSet()

    @DynamicGetattrSetter.register
    def depend(self):
        return self._generate_depset(atom, "DEPEND")

    @DynamicGetattrSetter.register
    def rdepend(self):
        return self._generate_depset(atom, "RDEPEND")

    @DynamicGetattrSetter.register
    def pdepend(self):
        return self._generate_depset(atom, "PDEPEND")

    @DynamicGetattrSetter.register
    def idepend(self):
        if "IDEPEND" in self.eapi.metadata_keys:
            return self._generate_depset(atom, "IDEPEND")
        return conditionals.DepSet()

    @DynamicGetattrSetter.register
    def license(self):
        return conditionals.DepSet.parse(
            self.data.pop('LICENSE', ''), str,
            operators={
                '||': boolean.OrRestriction,
                '': boolean.AndRestriction},
            attr='LICENSE', element_func=intern)

    @DynamicGetattrSetter.register
    def fullslot(self):
        slot = self.data.get('SLOT', None)
        if not slot:
            raise metadata_errors.MetadataException(
                self, 'slot', 'SLOT cannot be unset or empty')
        if not self.eapi.valid_slot_regex.match(slot):
            raise metadata_errors.MetadataException(
                self, 'slot', f'invalid SLOT: {slot!r}')
        return slot

    @DynamicGetattrSetter.register
    def subslot(self):
        slot, _sep, subslot = self.fullslot.partition('/')
        if not subslot:
            return slot
        return subslot

    @DynamicGetattrSetter.register
    def slot(self):
        return self.fullslot.partition('/')[0]

    def create_fetchable_from_uri(
            self, chksums, ignore_missing_chksums, ignore_unknown_mirrors,
            mirrors, default_mirrors, common_files, uri, filename=None):
        default_filename = os.path.basename(uri)
        if filename is not None:
            # log redundant renames for pkgcheck to flag
            if filename == default_filename:
                logger.info(f'redundant rename: {uri} -> {filename}')
        else:
            filename = default_filename

        if not filename:
            raise ValueError(f'missing filename: {uri!r}')

        preexisting = common_files.get(filename)

        if preexisting is None:
            if filename not in chksums and not ignore_missing_chksums:
                raise metadata_errors.MissingChksum(self, filename)
            uris = fetch.uri_list(filename)
        else:
            uris = preexisting.uri

        # fetch restriction implies mirror restriction
        pkg_allow_fetch = "fetch" not in self.restrict
        pkg_allow_mirror = "mirror" not in self.restrict and pkg_allow_fetch

        if filename != uri:
            unrestrict_mirror = unrestrict_fetch = False
            if self.eapi.options.src_uri_unrestrict:
                # mirror unrestriction implies fetch unrestriction
                unrestrict_mirror = uri.startswith('mirror+')
                unrestrict_fetch = uri.startswith('fetch+') or unrestrict_mirror
                if unrestrict_fetch:
                    # strip the prefix
                    uri = uri.partition('+')[2]

            allow_mirror = pkg_allow_mirror or unrestrict_mirror

            if preexisting is None:
                if "primaryuri" not in self.restrict:
                    if default_mirrors and allow_mirror:
                        uris.add_mirror(default_mirrors)

            if uri.startswith("mirror://"):
                # mirror:// is 9 chars.
                tier, remaining_uri = uri[9:].split("/", 1)
                mirror = mirrors.get(tier, fetch.unknown_mirror(tier))
                uris.add_mirror(mirror, sub_uri=remaining_uri)

            else:
                uris.add_uri(uri)
            if preexisting is None and "primaryuri" in self.restrict:
                if default_mirrors and allow_mirror:
                    uris.add_mirror(default_mirrors)

        if preexisting is None:
            common_files[filename] = fetch.fetchable(filename, uris, chksums.get(filename))
        return common_files[filename]

    def generate_fetchables(self, allow_missing_checksums=False,
                            ignore_unknown_mirrors=False, skip_default_mirrors=False):
        """Generate fetchables object for a package."""
        chksums_can_be_missing = allow_missing_checksums or \
            bool(getattr(self.repo, '_allow_missing_chksums', False))
        chksums_can_be_missing, chksums = self.repo._get_digests(
            self, allow_missing=chksums_can_be_missing)

        mirrors = getattr(self._parent, "mirrors", {})
        if skip_default_mirrors:
            default_mirrors = None
        else:
            default_mirrors = getattr(self._parent, "default_mirrors", None)
        common = {}
        func = partial(
            self.create_fetchable_from_uri, chksums,
            chksums_can_be_missing, ignore_unknown_mirrors,
            mirrors, default_mirrors, common)

        try:
            d = conditionals.DepSet.parse(
                self.data.get("SRC_URI", ""), fetch.fetchable, operators={},
                element_func=func, attr='SRC_URI',
                allow_src_uri_file_renames=self.eapi.options.src_uri_renames)
        except ebuild_errors.DepsetParseError as e:
            raise metadata_errors.MetadataException(self, 'fetchables', str(e))

        for v in common.values():
            v.uri.finalize()
        return d

    @DynamicGetattrSetter.register
    def fetchables(self):
        return self.generate_fetchables()

    @DynamicGetattrSetter.register
    def distfiles(self):
        def _extract_distfile_from_uri(uri, filename=None):
            if filename is not None:
                return filename
            return os.path.basename(uri)
        return conditionals.DepSet.parse(
            self.data.get("SRC_URI", ''), str, operators={}, attr='SRC_URI',
            element_func=partial(_extract_distfile_from_uri),
            allow_src_uri_file_renames=self.eapi.options.src_uri_renames)

    @DynamicGetattrSetter.register
    def description(self):
        return self.data.pop("DESCRIPTION", "").strip()

    @DynamicGetattrSetter.register
    def keywords(self):
        return tuple(map(intern, self.data.pop("KEYWORDS", "").split()))

    @property
    def sorted_keywords(self):
        """Sort keywords with prefix keywords after regular arches."""
        return tuple(sort_keywords(self.keywords))

    @DynamicGetattrSetter.register
    def restrict(self):
        return conditionals.DepSet.parse(
            self.data.pop("RESTRICT", ''), str, operators={},
            attr='RESTRICT')

    @DynamicGetattrSetter.register
    def eapi(self):
        ebuild = self.ebuild
        eapi = '0'
        if ebuild.path:
            # Use readlines directly since it does whitespace stripping
            # for us, far faster than native python can.
            i = fileutils.readlines_utf8(ebuild.path)
        else:
            i = (x.strip() for x in ebuild.text_fileobj())
        for line in i:
            if line[0:1] in ('', '#'):
                continue
            if (mo := _EAPI_str_regex.match(line)) and (eapi_str := mo.group('EAPI')):
                eapi = _EAPI_regex.match(line).group('EAPI')
            break
        i.close()
        try:
            return get_eapi(eapi)
        except ValueError as e:
            error = str(e) if eapi else f'{e}: {eapi_str!r}'
            raise metadata_errors.MetadataException(self, 'eapi', error)

    is_supported = klass.alias_attr('eapi.is_supported')
    tracked_attributes = klass.alias_attr('eapi.tracked_attributes')

    @DynamicGetattrSetter.register
    def iuse(self):
        return frozenset(map(intern, self.data.pop("IUSE", "").split()))

    @property
    def iuse_stripped(self):
        if self.eapi.options.iuse_defaults:
            return frozenset(x.lstrip('-+') if len(x) > 1 else x for x in self.iuse)
        return self.iuse

    iuse_effective = klass.alias_attr("iuse_stripped")

    @DynamicGetattrSetter.register
    def user_patches(self):
        return ()

    @DynamicGetattrSetter.register
    def properties(self):
        return conditionals.DepSet.parse(
            self.data.pop("PROPERTIES", ''), str, operators={},
            attr='PROPERTIES')

    @DynamicGetattrSetter.register
    def defined_phases(self):
        return self.eapi.interpret_cache_defined_phases(
            map(intern, self.data.pop("DEFINED_PHASES", "").split()))

    @DynamicGetattrSetter.register
    def homepage(self):
        return tuple(self.data.pop("HOMEPAGE", "").split())

    @DynamicGetattrSetter.register
    def inherited(self):
        """Ordered set of all inherited eclasses."""
        return OrderedFrozenSet(self.data.get("_eclasses_", ()))

    @DynamicGetattrSetter.register
    def inherit(self):
        """Ordered set of directly inherited eclasses."""
        return OrderedFrozenSet(self.data.get("INHERIT", "").split())

    @staticmethod
    def _mk_required_use_node(data):
        if data[0] == '!':
            return values.ContainmentMatch(data[1:], negate=True)
        return values.ContainmentMatch(data)

    @DynamicGetattrSetter.register
    def required_use(self):
        if self.eapi.options.has_required_use:
            data = self.data.pop("REQUIRED_USE", "")
            if data:
                operators = {
                    "||": boolean.OrRestriction,
                    "": boolean.AndRestriction,
                    "^^": boolean.JustOneRestriction
                }

                def _invalid_op(msg, *args):
                    raise metadata_errors.MetadataException(self, 'eapi', f'REQUIRED_USE: {msg}')

                if self.eapi.options.required_use_one_of:
                    operators['??'] = boolean.AtMostOneOfRestriction
                else:
                    operators['??'] = partial(
                        _invalid_op, f"EAPI '{self.eapi}' doesn't support '??' operator")

                return conditionals.DepSet.parse(
                    data,
                    values.ContainmentMatch, operators=operators,
                    element_func=self._mk_required_use_node, attr='REQUIRED_USE')
        return conditionals.DepSet()

    source_repository = klass.alias_attr("repo.repo_id")

    PN = klass.alias_attr("package")
    PV = klass.alias_attr("version")
    PVR = klass.alias_attr("fullver")

    @property
    def mandatory_phases(self):
        return frozenset(
            chain(self.defined_phases, self.eapi.default_phases))

    @property
    def live(self):
        return 'live' in self.properties

    @property
    def P(self):
        return f"{self.package}-{self.version}"

    @property
    def PF(self):
        return f"{self.package}-{self.fullver}"

    @property
    def PR(self):
        return f'r{self.revision}'

    @property
    def path(self):
        return self._parent._get_ebuild_path(self)

    @property
    def ebuild(self):
        return self._parent.get_ebuild_src(self)

    def _fetch_metadata(self, ebp=None, force_regen=None):
        return self._parent._get_metadata(self, ebp=ebp, force_regen=force_regen)

    def __str__(self):
        return f"ebuild src: {self.cpvstr}"

    def __repr__(self):
        return "<%s cpv=%r @%#8x>" % (self.__class__, self.cpvstr, id(self))


class package(base):

    __slots__ = ("_shared_pkg_data",)

    def __init__(self, shared_pkg_data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, "_shared_pkg_data", shared_pkg_data)

    maintainers = klass.alias_attr("_shared_pkg_data.metadata_xml.maintainers")
    upstreams = klass.alias_attr("_shared_pkg_data.metadata_xml.upstreams")
    local_use = klass.alias_attr("_shared_pkg_data.metadata_xml.local_use")
    longdescription = klass.alias_attr("_shared_pkg_data.metadata_xml.longdescription")
    manifest = klass.alias_attr("_shared_pkg_data.manifest")
    stabilize_allarches = klass.alias_attr("_shared_pkg_data.metadata_xml.stabilize_allarches")

    @property
    def _mtime_(self):
        return self._parent._get_ebuild_mtime(self)

    @property
    def environment(self):
        data = self._get_ebuild_environment()
        return data_source.data_source(data, mutable=False)

    def _get_ebuild_environment(self, ebp=None):
        with processor.reuse_or_request(ebp) as ebp:
            return ebp.get_ebuild_environment(self, self.repo.eclass_cache)


class package_factory(metadata.factory):

    child_class = package

    # For the plugin system.
    priority = 5

    def __init__(self, parent, cachedb, eclass_cache, mirrors, default_mirrors,
                 *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self._cache = cachedb
        self._ecache = eclass_cache

        if mirrors:
            mirrors = {k: fetch.mirror(v, k) for k, v in mirrors.items()}

        self.mirrors = mirrors
        if default_mirrors:
            self.default_mirrors = fetch.default_mirror(default_mirrors, "conf. default mirror")
        else:
            self.default_mirrors = None

    def get_ebuild_src(self, pkg):
        return self._parent_repo._get_ebuild_src(pkg)

    def _get_ebuild_path(self, pkg):
        return self._parent_repo._get_ebuild_path(pkg)

    def _get_ebuild_mtime(self, pkg):
        return os.stat(self._get_ebuild_path(pkg)).st_mtime

    def _get_metadata(self, pkg, ebp=None, force_regen=False):
        caches = self._cache
        if force_regen:
            caches = ()
        ebuild_hash = chksum.LazilyHashedPath(pkg.path)
        for cache in caches:
            if cache is not None:
                try:
                    data = cache[pkg.cpvstr]
                    if cache.validate_entry(data, ebuild_hash, self._ecache):
                        return data
                    if not cache.readonly:
                        del cache[pkg.cpvstr]
                except KeyError:
                    continue
                except cache_errors.CacheError as e:
                    logger.warning("caught cache error: %s", e)
                    del e
                    continue

        # no cache entries, regen
        return self._update_metadata(pkg, ebp=ebp)

    def _update_metadata(self, pkg, ebp=None):
        parsed_eapi = pkg.eapi
        if not parsed_eapi.is_supported:
            return {'EAPI': str(parsed_eapi)}

        with processor.reuse_or_request(ebp) as my_proc:
            try:
                mydata = my_proc.get_keys(pkg, self._ecache)
            except processor.ProcessorError as e:
                raise metadata_errors.MetadataException(
                    pkg, 'data', 'failed sourcing ebuild', e)

        # Rewrite defined_phases as needed, since we now know the EAPI.
        eapi = get_eapi(mydata.get('EAPI', '0'))
        if parsed_eapi != eapi:
            raise metadata_errors.MetadataException(
                pkg, 'eapi',
                f"parsed EAPI '{parsed_eapi}' doesn't match sourced EAPI '{eapi}'")
        wipes = set(mydata)

        wipes.difference_update(eapi.metadata_keys)
        if mydata["DEFINED_PHASES"] != '-':
            phases = mydata["DEFINED_PHASES"].split()
            d = eapi.phases_rev
            phases = set(d.get(x) for x in phases)
            # Discard is required should we have gotten
            # a phase that isn't actually in this EAPI.
            phases.discard(None)
            mydata["DEFINED_PHASES"] = ' '.join(sorted(phases))

        if inherited := mydata.pop("INHERITED", None):
            mydata["_eclasses_"] = self._ecache.get_eclass_data(inherited.split())
        mydata['_chf_'] = chksum.LazilyHashedPath(pkg.path)

        for x in wipes:
            del mydata[x]

        if self._cache is not None:
            for cache in self._cache:
                if not cache.readonly:
                    try:
                        cache[pkg.cpvstr] = mydata
                    except cache_errors.CacheError as e:
                        logger.warning("caught cache error: %s", e)
                        del e
                        continue
                    break

        return mydata

    def new_package(self, *args):
        if self._parent_repo.package_cache:
            inst = self._cached_instances.get(args)
        else:
            # package caching is disabled
            inst = None

        if inst is None:
            # key being cat/pkg
            mxml = self._parent_repo._get_shared_pkg_data(args[0], args[1])
            inst = self._cached_instances[args] = self.child_class(
                mxml, self, *args)

        return inst


generate_new_factory = package_factory



"""
default fs ops.

Shouldn't be accessed directly for the most part, use
:mod:`pkgcore.plugins` to get at these ops.
"""

import errno
import os
from functools import partial

from snakeoil.osutils import ensure_dirs, pjoin, unlink_if_exists
from snakeoil.process.spawn import spawn

from ..const import CP_BINARY
from . import contents, fs
from .livefs import gen_obj


def ensure_perms(d1, d2=None):

    """Enforce a fs objects attributes on the livefs.

    Attributes enforced are permissions, mtime, uid, gid.

    :param d2: if not None, an fs object for what's on the livefs now
    :return: True on success, else an exception is thrown
    :raise EnvironmentError: if fs object attributes can't be enforced
    """

    m, o, g, t = d1.mode, d1.uid, d1.gid, d1.mtime
    if o is None:
        o = -1
    if g is None:
        g = -1
    if d2 is None:
        do_mode, do_chown, do_mtime = True, True, True
    else:

        do_mode = False
        try:
            if fs.isdir(d1) and fs.isdir(d2):
                # if it's preexisting, keep its perms.
                do_mode = False
            else:
                do_mode = (m is not None and m != d2.mode)
        except AttributeError:
            # yes.  this _is_ stupid.  vdb's don't always store all attributes
            do_mode = False

        do_chown = False
        try:
            do_chown = (o != d2.uid or g != d2.gid)
        except AttributeError:
            do_chown = True

        try:
            do_mtime = (t != d2.mtime)
        except AttributeError:
            do_mtime = True

    if do_chown and (o != -1 or g != -1):
        os.lchown(d1.location, o, g)
    if not fs.issym(d1):
        if do_mode and m is not None:
            os.chmod(d1.location, m)
        if do_mtime and t is not None:
            os.utime(d1.location, (t, t))
    return True


def mkdir(d):
    """
    mkdir for a fsDir object

    :param d: :class:`pkgcore.fs.fs.fsDir` instance
    :return: true if success, else an exception is thrown
    :raise EnvironmentError: if can't complete
    """
    if not d.mode:
        mode = 0o777
    else:
        mode = d.mode
    os.mkdir(d.location, mode)
    ensure_perms(d)
    return True

# minor hack.

class FailedCopy(TypeError):

    def __init__(self, obj, msg):
        self.obj = obj
        self.msg = msg

    def __str__(self):
        return f'failed copying {self.obj}: {self.msg}'


class CannotOverwrite(FailedCopy):
    def __init__(self, obj, existing):
        self.obj, self.existing = obj, existing

    def __str__(self):
        return f'cannot write {self.obj} due to {self.existing} existing'


def copyfile(obj, mkdirs=False):
    """
    copy a :class:`pkgcore.fs.fs.fsBase` to its stated location.

    :param obj: :class:`pkgcore.fs.fs.fsBase` instance, exempting :class:`fsDir`
    :return: true if success, else an exception is thrown
    :raise EnvironmentError: permission errors

    """

    existent = False
    if not fs.isfs_obj(obj):
        raise TypeError(f'obj must be fsBase derivative: {obj!r}')
    elif fs.isdir(obj):
        raise TypeError(f'obj must not be a fsDir instance: {obj!r}')

    try:
        existing = gen_obj(obj.location)
        if fs.isdir(existing):
            raise CannotOverwrite(obj, existing)
        existent = True
    except OSError as oe:
        # verify the parent dir is there at least
        basefp = os.path.dirname(obj.location)
        if basefp.strip(os.path.sep) and not os.path.exists(basefp):
            if mkdirs:
                if not ensure_dirs(basefp, mode=0o750, minimal=True):
                    raise FailedCopy(obj, str(oe))
            else:
                raise
        existent = False

    if not existent:
        fp = obj.location
    else:
        fp = existent_fp = obj.location + "#new"

    if fs.isreg(obj):
        obj.data.transfer_to_path(fp)
    elif fs.issym(obj):
        os.symlink(obj.target, fp)
    elif fs.isfifo(obj):
        os.mkfifo(fp)
    elif fs.isdev(obj):
        dev = os.makedev(obj.major, obj.minor)
        os.mknod(fp, obj.mode, dev)
    else:
        ret = spawn([CP_BINARY, "-Rp", obj.location, fp])
        if ret != 0:
            raise FailedCopy(obj, f'got {ret} from {CP_BINARY} -Rp')

    ensure_perms(obj.change_attributes(location=fp))

    if existent:
        os.rename(existent_fp, obj.location)
    return True

def do_link(src, trg):
    try:
        os.link(src.location, trg.location)
        return True
    except FileExistsError:
        pass
    except EnvironmentError as e:
        if e.errno == errno.EXDEV:
            # hardlink is impossible, force copyfile
            return False
        raise

    path = trg.location + '#new'
    unlink_if_exists(path)
    try:
        os.link(src.location, path)
    except EnvironmentError as e:
        if e.errno != errno.EXDEV:
            # someone is screwing with us, or unlink_if_exists is broken.
            raise
        # hardlink is impossible, force copyfile
        return False
    try:
        os.rename(path, trg.location)
    except EnvironmentError:
        unlink_if_exists(path)
        if e.eerrno != errno.EXDEV:
            # weird error, broken FS codes, perms, or someone is screwing with us.
            raise
        # this is only possible on overlay fs's; while annoying, you can have two
        # different filesystems in use in the same directory in those cases.
        return False
    return True


def merge_contents(cset, offset=None, callback=None):

    """
    merge a :class:`pkgcore.fs.contents.contentsSet` instance to the livefs

    :param cset: :class:`pkgcore.fs.contents.contentsSet` instance
    :param offset: if not None, offset to prefix all locations with.
        Think of it as target dir.
    :param callback: callable to report each entry being merged; given a single arg,
        the fs object being merged.
    :raise EnvironmentError: Thrown for permission failures.
    """

    if callback is None:
        callback = lambda obj:None

    if not isinstance(cset, contents.contentsSet):
        raise TypeError(f'cset must be a contentsSet, got {cset!r}')

    if offset is not None:
        if os.path.exists(offset):
            if not os.path.isdir(offset):
                raise TypeError(f'offset must be a dir, or not exist: {offset}')
        else:
            mkdir(fs.fsDir(offset, strict=False))
        iterate = partial(contents.offset_rewriter, offset.rstrip(os.path.sep))
    else:
        iterate = iter

    d = list(iterate(cset.iterdirs()))
    d.sort()
    for x in d:
        callback(x)

        try:
            # we pass in the stat ourselves, using stat instead of
            # lstat gen_obj uses internally; this is the equivalent of
            # "deference that link"
            obj = gen_obj(x.location, stat=os.stat(x.location))
            if not fs.isdir(obj):
                # according to the spec, dirs can't be merged over files
                # that aren't dirs or symlinks to dirs
                raise CannotOverwrite(x.location, obj)
            ensure_perms(x, obj)
        except FileNotFoundError:
            try:
                # we do this form to catch dangling symlinks
                mkdir(x)
            except FileExistsError:
                os.unlink(x.location)
                mkdir(x)
            ensure_perms(x)
    del d

    # might look odd, but what this does is minimize the try/except cost
    # to one time, assuming everything behaves, rather then per item.
    i = iterate(cset.iterdirs(invert=True))
    merged_inodes = {}
    while True:
        try:
            for x in i:
                callback(x)

                if x.is_reg:
                    key = (x.dev, x.inode)
                    # This logic could be made smarter- instead of
                    # blindly trying candidates, we could inspect the st_dev
                    # of the final location.  This however can be broken by
                    # overlayfs's potentially.  Brute force is in use either
                    # way.
                    candidates = merged_inodes.setdefault(key, [])
                    if any(target._can_be_hardlinked(x) and do_link(target, x)
                            for target in candidates):
                        continue
                    candidates.append(x)

                copyfile(x, mkdirs=True)

            break
        except CannotOverwrite as cf:
            if not fs.issym(x):
                raise

            # by this time, all directories should've been merged.
            # thus we can check the target
            try:
                if not fs.isdir(gen_obj(pjoin(x.location, x.target))):
                    raise
            except OSError:
                raise cf
    return True


def unmerge_contents(cset, offset=None, callback=None):

    """
    unmerge a :obj:`pkgcore.fs.contents.contentsSet` instance to the livefs

    :param cset: :obj:`pkgcore.fs.contents.contentsSet` instance
    :param offset: if not None, offset to prefix all locations with.
        Think of it as target dir.
    :param callback: callable to report each entry being unmerged
    :return: True, or an exception is thrown on failure
        (OSError, although see copyfile for specifics).
    :raise EnvironmentError: see :func:`copyfile` and :func:`mkdir`
    """

    if callback is None:
        callback = lambda obj: None

    iterate = iter
    if offset is not None:
        iterate = partial(contents.offset_rewriter, offset.rstrip(os.path.sep))

    for x in iterate(cset.iterdirs(invert=True)):
        callback(x)
        unlink_if_exists(x.location)

    # this is a fair sight faster then using sorted/reversed
    l = list(iterate(cset.iterdirs()))
    l.sort(reverse=True)
    for x in l:
        try:
            os.rmdir(x.location)
        except OSError as e:
            if not e.errno in (errno.ENOTEMPTY, errno.ENOENT, errno.ENOTDIR,
                               errno.EBUSY, errno.EEXIST):
                raise
        else:
            callback(x)
    return True
